#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2024 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from datetime import datetime

try:
    import pyarrow as pa
except (AttributeError, ImportError):
    pa = None

from .. import types as odps_types, serializers, utils, readers
from ..compat import dir2, six, Enum
from ..config import options
from .cluster_info import ClusterInfo, ClusterType
from .core import LazyLoad, JSONRemoteModel
from .partitions import Partitions
from .record import Record
from .storage_tier import StorageTier, StorageTierInfo
from .tableio import (
    TableArrowReader,
    TableArrowWriter,
    TableRecordReader,
    TableRecordWriter,
)


class TableSchema(odps_types.OdpsSchema, JSONRemoteModel):
    """
    Schema includes the columns and partitions information of a :class:`odps.models.Table`.

    There are two ways to initialize a Schema object, first is to provide columns and partitions,
    the second way is to call the class method ``from_lists``. See the examples below:

    :Example:

    >>> columns = [Column(name='num', type='bigint', comment='the column')]
    >>> partitions = [Partition(name='pt', type='string', comment='the partition')]
    >>> schema = TableSchema(columns=columns, partitions=partitions)
    >>> schema.columns
    [<column num, type bigint>, <partition pt, type string>]
    >>>
    >>> schema = TableSchema.from_lists(['num'], ['bigint'], ['pt'], ['string'])
    >>> schema.columns
    [<column num, type bigint>, <partition pt, type string>]
    """

    class Shard(JSONRemoteModel):

        hub_lifecycle = serializers.JSONNodeField('HubLifecycle')
        shard_num = serializers.JSONNodeField('ShardNum')
        distribute_cols = serializers.JSONNodeField('DistributeCols')
        sort_cols = serializers.JSONNodeField('SortCols')

    class TableColumn(odps_types.Column, JSONRemoteModel):
        name = serializers.JSONNodeField('name')
        type = serializers.JSONNodeField('type', parse_callback=odps_types.validate_data_type)
        comment = serializers.JSONNodeField('comment')
        label = serializers.JSONNodeField('label')
        nullable = serializers.JSONNodeField('isNullable')

        def __init__(self, **kwargs):
            kwargs.setdefault("nullable", True)
            JSONRemoteModel.__init__(self, **kwargs)
            if self.type is not None:
                self.type = odps_types.validate_data_type(self.type)

    class TablePartition(odps_types.Partition, TableColumn):
        def __init__(self, **kwargs):
            TableSchema.TableColumn.__init__(self, **kwargs)

    def __init__(self, **kwargs):
        kwargs['_columns'] = columns = kwargs.pop('columns', None)
        kwargs['_partitions'] = partitions = kwargs.pop('partitions', None)
        JSONRemoteModel.__init__(self, **kwargs)
        odps_types.OdpsSchema.__init__(self, columns=columns, partitions=partitions)

    def load(self):
        self.update(self._columns, self._partitions)
        self.build_snapshot()

    comment = serializers.JSONNodeField('comment', set_to_parent=True)
    owner = serializers.JSONNodeField('owner', set_to_parent=True)
    creation_time = serializers.JSONNodeField(
        'createTime', parse_callback=datetime.fromtimestamp,
        set_to_parent=True)
    last_data_modified_time = serializers.JSONNodeField(
        'lastModifiedTime', parse_callback=datetime.fromtimestamp,
        set_to_parent=True)
    last_meta_modified_time = serializers.JSONNodeField(
        'lastDDLTime', parse_callback=datetime.fromtimestamp,
        set_to_parent=True)
    is_virtual_view = serializers.JSONNodeField(
        'isVirtualView', parse_callback=bool, set_to_parent=True)
    is_materialized_view = serializers.JSONNodeField(
        'isMaterializedView', parse_callback=bool, set_to_parent=True)
    is_materialized_view_rewrite_enabled = serializers.JSONNodeField(
        'isMaterializedViewRewriteEnabled',
        parse_callback=lambda x: x is not None and str(x).lower() == "true",
        set_to_parent=True,
    )
    is_materialized_view_outdated = serializers.JSONNodeField(
        'isMaterializedViewOutdated',
        parse_callback=lambda x: x is not None and str(x).lower() == "true",
        set_to_parent=True,
    )
    lifecycle = serializers.JSONNodeField(
        'lifecycle', parse_callback=int, set_to_parent=True)
    view_text = serializers.JSONNodeField('viewText', set_to_parent=True)
    view_expanded_text = serializers.JSONNodeField('viewExpandedText', set_to_parent=True)
    size = serializers.JSONNodeField("size", parse_callback=int, set_to_parent=True)
    is_archived = serializers.JSONNodeField(
        'IsArchived', parse_callback=bool, set_to_parent=True)
    physical_size = serializers.JSONNodeField(
        'PhysicalSize', parse_callback=int, set_to_parent=True)
    file_num = serializers.JSONNodeField(
        'FileNum', parse_callback=int, set_to_parent=True)
    record_num = serializers.JSONNodeField(
        'recordNum', parse_callback=int, set_to_parent=True)
    location = serializers.JSONNodeField(
        'location', set_to_parent=True)
    storage_handler = serializers.JSONNodeField(
        'storageHandler', set_to_parent=True)
    resources = serializers.JSONNodeField(
        'resources', set_to_parent=True)
    serde_properties = serializers.JSONNodeField(
        'serDeProperties', type='json', set_to_parent=True)
    reserved = serializers.JSONNodeField(
        'Reserved', type='json', set_to_parent=True)
    shard = serializers.JSONNodeReferenceField(
        Shard, 'shardInfo', check_before=['shardExist'], set_to_parent=True)
    table_label = serializers.JSONNodeField(
        'tableLabel', callback=lambda t: t if t != '0' else '', set_to_parent=True)
    _columns = serializers.JSONNodesReferencesField(TableColumn, 'columns')
    _partitions = serializers.JSONNodesReferencesField(TablePartition, 'partitionKeys')

    def __getstate__(self):
        return self._columns, self._partitions

    def __setstate__(self, state):
        columns, partitions = state
        self.__init__(columns=columns, partitions=partitions)

    def __dir__(self):
        return sorted(set(dir2(self)) - set(type(self)._parent_attrs))


class Table(LazyLoad):
    """
    Table means the same to the RDBMS table, besides, a table can consist of partitions.

    Table's properties are the same to the ones of :class:`odps.models.Project`,
    which will not load from remote ODPS service until users try to get them.

    In order to write data into table, users should call the ``open_writer``
    method with **with statement**. At the same time, the ``open_reader`` method is used
    to provide the ability to read records from a table or its partition.

    :Example:

    >>> table = odps.get_table('my_table')
    >>> table.owner  # first will load from remote
    >>> table.reload()  # reload to update the properties
    >>>
    >>> for record in table.head(5):
    >>>     # check the first 5 records
    >>> for record in table.head(5, partition='pt=test', columns=['my_column'])
    >>>     # only check the `my_column` column from certain partition of this table
    >>>
    >>> with table.open_reader() as reader:
    >>>     count = reader.count  # How many records of a table or its partition
    >>>     for record in reader[0: count]:
    >>>         # read all data, actually better to split into reading for many times
    >>>
    >>> with table.open_writer() as writer:
    >>>     writer.write(records)
    >>> with table.open_writer(partition='pt=test', blocks=[0, 1]):
    >>>     writer.write(0, gen_records(block=0))
    >>>     writer.write(1, gen_records(block=1))  # we can do this parallel
    """
    _extend_args = (
        'is_archived', 'physical_size', 'file_num', 'location', 'storage_handler',
        'resources', 'serde_properties', 'reserved', 'is_transactional',
        'primary_key', 'storage_tier_info', 'cluster_info'
    )
    __slots__ = (
        '_is_extend_info_loaded', 'last_meta_modified_time', 'is_virtual_view',
        'is_materialized_view', 'is_materialized_view_rewrite_enabled',
        'is_materialized_view_outdated', 'lifecycle', 'view_text',
        'view_expanded_text', 'size', 'shard', 'record_num',
        '_table_tunnel', '_id_thread_local'
    )
    __slots__ += _extend_args

    class Type(Enum):
        MANAGED_TABLE = "MANAGED_TABLE"
        VIRTUAL_VIEW = "VIRTUAL_VIEW"
        EXTERNAL_TABLE = "EXTERNAL_TABLE"
        MATERIALIZED_VIEW = "MATERIALIZED_VIEW"

    name = serializers.XMLNodeField('Name')
    table_id = serializers.XMLNodeField('TableId')
    format = serializers.XMLNodeAttributeField(attr='format')
    table_schema = serializers.XMLNodeReferenceField(TableSchema, 'Schema')
    comment = serializers.XMLNodeField('Comment')
    owner = serializers.XMLNodeField('Owner')
    table_label = serializers.XMLNodeField('TableLabel')
    creation_time = serializers.XMLNodeField(
        'CreationTime', parse_callback=utils.parse_rfc822
    )
    last_data_modified_time = serializers.XMLNodeField(
        'LastModifiedTime', parse_callback=utils.parse_rfc822
    )
    last_access_time = serializers.XMLNodeField(
        'LastAccessTime', parse_callback=utils.parse_rfc822
    )
    type = serializers.XMLNodeField(
        'Type', parse_callback=lambda s: Table.Type(s.upper()) if s is not None else None
    )

    _download_ids = utils.thread_local_attribute('_id_thread_local', dict)
    _upload_ids = utils.thread_local_attribute('_id_thread_local', dict)

    def __init__(self, **kwargs):
        self._is_extend_info_loaded = False
        if 'schema' in kwargs:
            warnings.warn(
                'Argument schema is deprecated and will be replaced by table_schema.',
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs['table_schema'] = kwargs.pop('schema')
        super(Table, self).__init__(**kwargs)

        try:
            del self._id_thread_local
        except AttributeError:
            pass

    def table_resource(self, client=None, endpoint=None, force_schema=False):
        schema_name = self._get_schema_name()
        if force_schema:
            schema_name = schema_name or "default"
        if schema_name is None:
            return self.resource(client=client, endpoint=endpoint)
        return "/".join(
            [
                self.project.resource(client, endpoint=endpoint),
                "schemas",
                schema_name,
                "tables",
                self.name,
            ]
        )

    @property
    def full_table_name(self):
        schema_name = self._get_schema_name()
        if schema_name is None:
            return '{0}.`{1}`'.format(self.project.name, self.name)
        else:
            return '{0}.{1}.`{2}`'.format(self.project.name, schema_name, self.name)

    def reload(self):
        url = self.resource()
        resp = self._client.get(url, curr_schema=self._get_schema_name())

        self.parse(self._client, resp, obj=self)
        self.table_schema.load()
        self._loaded = True

    def reset(self):
        super(Table, self).reset()
        self._is_extend_info_loaded = False
        self.table_schema = None

    @property
    def schema(self):
        warnings.warn(
            'Table.schema is deprecated and will be replaced by Table.table_schema.',
            DeprecationWarning,
            stacklevel=3,
        )
        utils.add_survey_call(".".join([type(self).__module__, type(self).__name__, "schema"]))
        return self.table_schema

    @property
    def last_modified_time(self):
        warnings.warn(
            "Table.last_modified_time is deprecated and will be replaced by "
            "Table.last_data_modified_time.",
            DeprecationWarning,
            stacklevel=3,
        )
        utils.add_survey_call(".".join(
            [type(self).__module__, type(self).__name__, "last_modified_time"]
        ))
        return self.last_data_modified_time

    def _parse_reserved(self):
        if not self.reserved:
            self.is_transactional = None
            self.primary_key = None
            self.storage_tier_info = None
            self.cluster_info = None
            return
        is_transactional = self.reserved.get("Transactional")
        self.is_transactional = (
            is_transactional is not None and is_transactional.lower() == "true"
        )
        self.primary_key = self.reserved.get('PrimaryKey')
        self.storage_tier_info = StorageTierInfo.deserial(self.reserved)
        self.cluster_info = ClusterInfo.deserial(self.reserved)

    def reload_extend_info(self):
        params = {}
        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['curr_schema'] = schema_name
        resp = self._client.get(self.resource(), action='extended', params=params)

        self.parse(self._client, resp, obj=self)
        self._is_extend_info_loaded = True

        if not self._loaded:
            self.table_schema = None

        self._parse_reserved()

    def __getattribute__(self, attr):
        if attr in type(self)._extend_args:
            if not self._is_extend_info_loaded:
                self.reload_extend_info()
                return object.__getattribute__(self, attr)

        val = object.__getattribute__(self, attr)
        if val is None and not self._loaded:
            if attr in getattr(TableSchema, '__fields'):
                self.reload()
                return object.__getattribute__(self, attr)

        return super(Table, self).__getattribute__(attr)

    def _repr(self):
        buf = six.StringIO()

        buf.write('odps.Table\n')
        buf.write('  name: {0}\n'.format(self.full_table_name))
        if self.type:
            buf.write('  type: {0}\n'.format(self.type.value))

        name_space = 2 * max(len(col.name) for col in self.table_schema.columns)
        type_space = 2 * max(len(repr(col.type)) for col in self.table_schema.columns)

        not_empty = lambda field: field is not None and len(field.strip()) > 0

        buf.write('  schema:\n')
        cols_strs = []
        for col in self.table_schema._columns:
            cols_strs.append('{0}: {1}{2}'.format(
                col.name.ljust(name_space),
                repr(col.type).ljust(type_space),
                '# {0}'.format(utils.to_str(col.comment)) if not_empty(col.comment) else ''
            ))
        buf.write(utils.indent('\n'.join(cols_strs), 4))
        buf.write('\n')

        if self.table_schema.partitions:
            buf.write('  partitions:\n')

            partition_strs = []
            for partition in self.table_schema.partitions:
                partition_strs.append('{0}: {1}{2}'.format(
                    partition.name.ljust(name_space),
                    repr(partition.type).ljust(type_space),
                    '# {0}'.format(utils.to_str(partition.comment)) if not_empty(partition.comment) else ''
                ))
            buf.write(utils.indent('\n'.join(partition_strs), 4))

        if self.view_text:
            buf.write('  view text:\n{0}'.format(utils.indent(self.view_text, 4)))

        return buf.getvalue()

    @property
    def stored_as(self):
        return (self.reserved or dict()).get('StoredAs')

    @classmethod
    def gen_create_table_sql(
        cls, table_name, table_schema, comment=None, if_not_exists=False,
        lifecycle=None, shard_num=None, hub_lifecycle=None,
        with_column_comments=True, transactional=False, primary_key=None,
        storage_tier=None, project=None, schema=None, table_type=None,
        view_text=None, **kw
    ):
        from ..utils import escape_odps_string

        buf = six.StringIO()
        table_name = utils.to_text(table_name)
        project = utils.to_text(project)
        schema = utils.to_text(schema)
        comment = utils.to_text(comment)
        view_text = utils.to_text(view_text)
        table_type = cls.Type(table_type or cls.Type.MANAGED_TABLE)
        is_view = table_type in (cls.Type.VIRTUAL_VIEW, cls.Type.MATERIALIZED_VIEW)
        primary_key = [primary_key] if isinstance(primary_key, six.string_types) else primary_key

        stored_as = kw.get('stored_as')
        external_stored_as = kw.get('external_stored_as')
        storage_handler = kw.get('storage_handler')
        table_properties = kw.get("table_properties") or {}
        cluster_info = kw.get('cluster_info')

        rewrite_enabled = kw.get('rewrite_enabled')
        rewrite_enabled = rewrite_enabled if rewrite_enabled is not None else True

        if table_type == cls.Type.EXTERNAL_TABLE:
            type_str = u'EXTERNAL TABLE'
        elif table_type == cls.Type.VIRTUAL_VIEW:
            type_str = u'VIEW'
        elif table_type == cls.Type.MATERIALIZED_VIEW:
            type_str = u'MATERIALIZED VIEW'
        else:
            type_str = u'EXTERNAL TABLE' if storage_handler or external_stored_as else u'TABLE'

        buf.write(u'CREATE %s ' % type_str)
        if if_not_exists:
            buf.write(u'IF NOT EXISTS ')
        if project is not None:
            buf.write(u'%s.' % project)
        if schema is not None:
            buf.write(u'%s.' % schema)
        buf.write(u'`%s` ' % table_name)

        if is_view and lifecycle is not None and lifecycle > 0:
            buf.write("LIFECYCLE %s " % lifecycle)

        def _write_primary_key(prev=""):
            if not primary_key:
                return
            if not prev.strip().endswith(","):
                buf.write(u",\n")
            buf.write(u"  PRIMARY KEY (%s)" % ", ".join("`%s`" % c for c in primary_key))

        if isinstance(table_schema, six.string_types):
            buf.write(u'(\n')
            buf.write(table_schema)
            _write_primary_key(table_schema)
            buf.write(u'\n)\n')
            if comment:
                buf.write(u"COMMENT '%s'\n" % escape_odps_string(comment))
        elif isinstance(table_schema, tuple):
            buf.write(u'(\n')
            buf.write(table_schema[0])
            _write_primary_key(table_schema[0])
            buf.write(u'\n)\n')
            if comment:
                buf.write(u"COMMENT '%s'\n" % escape_odps_string(comment))
            buf.write(u'PARTITIONED BY ')
            buf.write(u'(\n')
            buf.write(table_schema[1])
            buf.write(u'\n)\n')
        else:
            def write_columns(col_array, with_pk=False):
                size = len(col_array)
                buf.write(u'(\n')
                for idx, column in enumerate(col_array):
                    buf.write(u'  `%s` %s' % (utils.to_text(column.name), utils.to_text(column.type)))
                    if not column.nullable and not options.sql.ignore_fields_not_null:
                        buf.write(u" NOT NULL")
                    if with_column_comments and column.comment:
                        buf.write(u" COMMENT '%s'" % utils.to_text(column.comment))
                    if idx < size - 1:
                        buf.write(u',\n')
                if with_pk:
                    _write_primary_key()
                buf.write(u'\n)\n')

            def write_view_columns(col_array):
                size = len(col_array)
                buf.write(u'(\n')
                for idx, column in enumerate(col_array):
                    buf.write(u'  `%s`' % (utils.to_text(column.name)))
                    if with_column_comments and column.comment:
                        buf.write(u" COMMENT '%s'" % utils.to_text(column.comment))
                    if idx < size - 1:
                        buf.write(u',\n')
                buf.write(u'\n)\n')

            if not is_view:
                write_columns(table_schema.simple_columns, with_pk=True)
            else:
                write_view_columns(table_schema.simple_columns)

            if comment:
                buf.write(u"COMMENT '%s'\n" % comment)
            if table_type == cls.Type.MATERIALIZED_VIEW and not rewrite_enabled:
                buf.write(u'DISABLE REWRITE\n')
            if table_schema.partitions:
                if not is_view:
                    buf.write(u'PARTITIONED BY ')
                    write_columns(table_schema.partitions)
                else:
                    buf.write(u'PARTITIONED ON ')
                    write_view_columns(table_schema.partitions)

        if cluster_info is not None:
            if cluster_info.cluster_type == ClusterType.RANGE:
                cluster_type_str = u"RANGE "
            else:
                cluster_type_str = u""
            cluster_cols = u", ".join(u"`%s`" % col for col in cluster_info.cluster_cols)
            buf.write("%sCLUSTERED BY (%s)" % (cluster_type_str, cluster_cols))
            if cluster_info.sort_cols:
                sort_cols = u", ".join(
                    u"`%s` %s" % (c.name, c.order.value) for c in cluster_info.sort_cols
                )
                buf.write(u" SORTED BY (%s)" % sort_cols)
            if cluster_info.bucket_num:
                buf.write(" INTO %s BUCKETS" % cluster_info.bucket_num)
            buf.write(u"\n")

        if transactional:
            table_properties["transactional"] = "true"
        if storage_tier:
            if isinstance(storage_tier, six.string_types):
                storage_tier = StorageTier(utils.underline_to_camel(storage_tier).lower())
            table_properties["storagetier"] = storage_tier.value

        if table_properties:
            buf.write(u"TBLPROPERTIES (\n")
            for k, v in table_properties.items():
                buf.write(u'  "%s"="%s"' % (k, v))
            buf.write(u'\n)\n')

        serde_properties = kw.get('serde_properties')
        location = kw.get('location')
        resources = kw.get('resources')
        if storage_handler or external_stored_as:
            if storage_handler:
                buf.write("STORED BY '%s'\n" % escape_odps_string(storage_handler))
            else:
                buf.write("STORED AS %s\n" % escape_odps_string(external_stored_as))
            if serde_properties:
                buf.write('WITH SERDEPROPERTIES (\n')
                for idx, k in enumerate(serde_properties):
                    buf.write("  '%s' = '%s'" % (escape_odps_string(k), escape_odps_string(serde_properties[k])))
                    if idx + 1 < len(serde_properties):
                        buf.write(',')
                    buf.write('\n')
                buf.write(')\n')
            if location:
                buf.write("LOCATION '%s'\n" % location)
            if resources:
                buf.write("USING '%s'\n" % resources)
        if stored_as:
            buf.write("STORED AS %s\n" % escape_odps_string(stored_as))
        if not is_view and lifecycle is not None and lifecycle > 0:
            buf.write(u'LIFECYCLE %s\n' % lifecycle)
        if shard_num is not None:
            buf.write(u'INTO %s SHARDS' % shard_num)
            if hub_lifecycle is not None:
                buf.write(u' HUBLIFECYCLE %s\n' % hub_lifecycle)
            else:
                buf.write(u'\n')

        if is_view and view_text:
            buf.write(u'AS %s\n' % view_text)
        return buf.getvalue().strip()

    def get_ddl(self, with_comments=True, if_not_exists=False, force_table_ddl=False):
        """
        Get DDL SQL statement for the given table.

        :param with_comments: append comment for table and each column
        :param if_not_exists: generate `if not exists` code for generated DDL
        :param force_table_ddl: force generate table DDL if object is a view
        :return: DDL statement
        """
        shard_num = self.shard.shard_num if self.shard is not None else None
        storage_tier = (
            self.storage_tier_info.storage_tier.value if self.storage_tier_info else None
        )
        table_type = self.type if not force_table_ddl else self.Type.MANAGED_TABLE
        return self.gen_create_table_sql(
            self.name, self.table_schema, self.comment if with_comments else None,
            if_not_exists=if_not_exists, with_column_comments=with_comments,
            lifecycle=self.lifecycle, shard_num=shard_num, project=self.project.name,
            storage_handler=self.storage_handler, serde_properties=self.serde_properties,
            location=self.location, resources=self.resources, table_type=table_type,
            storage_tier=storage_tier, cluster_info=self.cluster_info,
            transactional=self.is_transactional, primary_key=self.primary_key,
            view_text=self.view_text,
            rewrite_enabled=self.is_materialized_view_rewrite_enabled,
        )

    @utils.survey
    def _head_by_data(self, limit, partition=None, columns=None, timeout=None):
        if limit <= 0:
            raise ValueError('limit number should >= 0.')

        params = {'linenum': limit}
        if partition is not None:
            if not isinstance(partition, odps_types.PartitionSpec):
                partition = odps_types.PartitionSpec(partition)
            params['partition'] = str(partition)
        if columns is not None and len(columns) > 0:
            col_name = lambda col: col.name if isinstance(col, odps_types.Column) else col
            params['cols'] = ','.join(col_name(col) for col in columns)

        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['schema_name'] = schema_name

        resp = self._client.get(
            self.resource(), action='data', params=params, stream=True, timeout=timeout
        )
        return readers.CsvRecordReader(self.table_schema, resp)

    def _head_by_preview(self, limit, partition=None, columns=None, compress_algo=None, timeout=None):
        table_tunnel = self._create_table_tunnel()
        return table_tunnel.open_preview_reader(
            self, partition_spec=partition, columns=columns, limit=limit,
            compress_algo=compress_algo, arrow=False, timeout=timeout, read_all=True
        )

    def head(self, limit, partition=None, columns=None, use_legacy=True, timeout=None):
        """
        Get the head records of a table or its partition.

        :param int limit: records' size, 10000 at most
        :param partition: partition of this table
        :param list columns: the columns which is subset of the table columns
        :return: records
        :rtype: list

        .. seealso:: :class:`odps.models.Record`
        """
        try:
            if pa is not None and not use_legacy:
                timeout = timeout if timeout is not None else options.tunnel.legacy_fallback_timeout
                return self._head_by_preview(
                    limit, partition=partition, columns=columns, timeout=timeout
                )
        except:
            # only raises when under tests and
            # use_legacy specified explicitly as False
            if use_legacy is False:
                raise
        return self._head_by_data(
            limit, partition=partition, columns=columns, timeout=timeout
        )

    def _create_table_tunnel(self, endpoint=None, quota_name=None):
        if self._table_tunnel is not None:
            return self._table_tunnel

        from ..tunnel import TableTunnel

        self._table_tunnel = TableTunnel(
            client=self._client,
            project=self.project,
            endpoint=endpoint or self.project._tunnel_endpoint,
            quota_name=quota_name,
        )
        return self._table_tunnel

    def open_reader(
        self,
        partition=None,
        reopen=False,
        endpoint=None,
        download_id=None,
        timeout=None,
        arrow=False,
        columns=None,
        quota_name=None,
        async_mode=True,
        **kw
    ):
        """
        Open the reader to read the entire records from this table or its partition.

        :param partition: partition of this table
        :param reopen: the reader will reuse last one, reopen is true means open a new reader.
        :type reopen: bool
        :param endpoint: the tunnel service URL
        :param download_id: use existing download_id to download table contents
        :param arrow: use arrow tunnel to read data
        :param quota_name: name of tunnel quota
        :param async_mode: enable async mode to create tunnels, can set True if session creation
            takes a long time.
        :param compress_option: compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :param compress_algo: compression algorithm, work when ``compress_option`` is not provided,
                              can be ``zlib``, ``snappy``
        :param compress_level: used for ``zlib``, work when ``compress_option`` is not provided
        :param compress_strategy: used for ``zlib``, work when ``compress_option`` is not provided
        :return: reader, ``count`` means the full size, ``status`` means the tunnel status

        :Example:

        >>> with table.open_reader() as reader:
        >>>     count = reader.count  # How many records of a table or its partition
        >>>     for record in reader[0: count]:
        >>>         # read all data, actually better to split into reading for many times
        """

        from ..tunnel.tabletunnel import TableDownloadSession

        if partition and not isinstance(partition, odps_types.PartitionSpec):
            partition = odps_types.PartitionSpec(partition)
        tunnel = self._create_table_tunnel(endpoint=endpoint, quota_name=quota_name)
        download_ids = dict()
        if download_id is None:
            download_ids = self._download_ids
            download_id = download_ids.get(partition) if not reopen else None
        download_session = utils.call_with_retry(
            tunnel.create_download_session,
            table=self, partition_spec=partition, download_id=download_id,
            timeout=timeout, async_mode=async_mode, **kw
        )

        if download_id and download_session.status != TableDownloadSession.Status.Normal:
            download_session = utils.call_with_retry(
                tunnel.create_download_session,
                table=self, partition_spec=partition, timeout=timeout,
                async_mode=async_mode, **kw
            )
        download_ids[partition] = download_session.id

        if arrow:
            return TableArrowReader(self, download_session, columns=columns)
        else:
            return TableRecordReader(self, download_session, partition, columns=columns)

    def open_writer(
        self,
        partition=None,
        blocks=None,
        reopen=False,
        create_partition=False,
        commit=True,
        endpoint=None,
        upload_id=None,
        arrow=False,
        quota_name=None,
        mp_context=None,
        **kw
    ):
        """
        Open the writer to write records into this table or its partition.

        :param partition: partition of this table
        :param blocks: block ids to open
        :param reopen: the reader will reuse last one, reopen is true means open a new reader.
        :type reopen: bool
        :param create_partition: if true, the partition will be created if not exist
        :type create_partition: bool
        :param endpoint: the tunnel service URL
        :param upload_id: use existing upload_id to upload data
        :param arrow: use arrow tunnel to write data
        :param quota_name: name of tunnel quota
        :param bool overwrite: if True, will overwrite existing data
        :param compress_option: compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :param compress_algo: compression algorithm, work when ``compress_option`` is not provided,
                              can be ``zlib``, ``snappy``
        :param compress_level: used for ``zlib``, work when ``compress_option`` is not provided
        :param compress_strategy: used for ``zlib``, work when ``compress_option`` is not provided
        :return: writer, status means the tunnel writer status

        :Example:

        >>> with table.open_writer() as writer:
        >>>     writer.write(records)
        >>> with table.open_writer(partition='pt=test', blocks=[0, 1]):
        >>>     writer.write(0, gen_records(block=0))
        >>>     writer.write(1, gen_records(block=1))  # we can do this parallel
        """

        from ..tunnel.tabletunnel import TableUploadSession

        if partition and not isinstance(partition, odps_types.PartitionSpec):
            partition = odps_types.PartitionSpec(partition)
        if create_partition and not self.exist_partition(create_partition):
            self.create_partition(partition, if_not_exists=True)

        tunnel = self._create_table_tunnel(endpoint=endpoint, quota_name=quota_name)
        if upload_id is None:
            upload_ids = self._upload_ids
            upload_id = upload_ids.get(partition) if not reopen else None
        else:
            upload_ids = dict()
        upload_session = utils.call_with_retry(
            tunnel.create_upload_session,
            table=self, partition_spec=partition, upload_id=upload_id, **kw
        )

        if upload_id and upload_session.status.value != TableUploadSession.Status.Normal.value:
            # check upload session status
            upload_session = utils.call_with_retry(
                tunnel.create_upload_session, table=self, partition_spec=partition, **kw
            )
            upload_id = None
        upload_ids[partition] = upload_session.id
        # as data of partition changed, remove existing download id to avoid TableModified error
        self._download_ids.pop(partition, None)

        writer_cls = TableArrowWriter if arrow else TableRecordWriter

        def _writer_on_close():
            if commit:
                upload_ids[partition] = None

        return writer_cls(
            self, upload_session, blocks, commit, on_close=_writer_on_close, mp_context=mp_context
        )

    @property
    def partitions(self):
        return Partitions(parent=self, client=self._client)

    @utils.with_wait_argument
    def create_partition(self, partition_spec, if_not_exists=False, async_=False, hints=None):
        """
        Create a partition within the table.

        :param partition_spec: specification of the partition.
        :param if_not_exists:
        :param hints:
        :param async_:
        :return: partition object
        :rtype: odps.models.partition.Partition
        """
        return self.partitions.create(
            partition_spec, if_not_exists=if_not_exists, hints=hints, async_=async_
        )

    @utils.with_wait_argument
    def delete_partition(self, partition_spec, if_exists=False, async_=False, hints=None):
        """
        Delete a partition within the table.

        :param partition_spec: specification of the partition.
        :param if_exists:
        :param hints:
        :param async_:
        """
        return self.partitions.delete(
            partition_spec, if_exists=if_exists, hints=hints, async_=async_
        )

    def exist_partition(self, partition_spec):
        """
        Check if a partition exists within the table.

        :param partition_spec: specification of the partition.
        """
        return partition_spec in self.partitions

    def exist_partitions(self, prefix_spec=None):
        """
        Check if partitions with provided conditions exist.

        :param prefix_spec: prefix of partition
        :return: whether partitions exist
        """
        try:
            next(self.partitions.iterate_partitions(spec=prefix_spec))
        except StopIteration:
            return False
        return True

    def iterate_partitions(self, spec=None, reverse=False):
        """
        Create an iterable object to iterate over partitions.

        :param spec: specification of the partition.
        :param reverse: output partitions in reversed order
        """
        return self.partitions.iterate_partitions(spec=spec, reverse=reverse)

    def get_partition(self, partition_spec):
        """
        Get a partition with given specifications.

        :param partition_spec: specification of the partition.
        :return: partition object
        :rtype: odps.models.partition.Partition
        """
        return self.partitions[partition_spec]

    def get_max_partition(self, spec=None, skip_empty=True, reverse=False):
        """
        Get partition with maximal values within certain spec.

        :param spec: parent partitions. if specified, will return partition with
            maximal value within specified parent partition
        :param skip_empty: if True, will skip partitions without data
        :param reverse: if True, will return minimal value
        :return: Partition
        """
        if not self.table_schema.partitions:
            raise ValueError("Table %r not partitioned" % self.name)
        return self.partitions.get_max_partition(
            spec, skip_empty=skip_empty, reverse=reverse
        )

    @utils.with_wait_argument
    def truncate(self, partition_spec=None, async_=False, hints=None):
        """
        truncate this table.

        :param partition_spec: partition specs
        :param hints:
        :param async_: run asynchronously if True
        :return: None
        """
        from .tasks import SQLTask

        partition_expr = ""
        if partition_spec is not None:
            if not isinstance(partition_spec, (list, tuple)):
                partition_spec = [partition_spec]
            partition_expr = " " + ", ".join(
                "PARTITION (%s)" % spec for spec in partition_spec
            )

        # as data of partition changed, remove existing download id to avoid TableModified error
        for part in (partition_spec or [None]):
            if isinstance(part, six.string_types):
                part = odps_types.PartitionSpec(part)
            self._download_ids.pop(part, None)

        task = SQLTask(
            name='SQLTruncateTableTask',
            query='TRUNCATE TABLE %s%s;' % (self.full_table_name, partition_expr)
        )

        hints = hints or {}
        hints['odps.sql.submit.mode'] = ''
        schema_name = self._get_schema_name()
        if schema_name is not None:
            hints["odps.sql.allow.namespace.schema"] = "true"
            hints["odps.namespace.schema"] = "true"
        task.update_sql_settings(hints)

        instance = self.project.parent[self._client.project].instances.create(task=task)

        if not async_:
            instance.wait_for_success()
        else:
            return instance

    @utils.with_wait_argument
    def drop(self, async_=False, if_exists=False, hints=None):
        """
        Drop this table.

        :param async_: run asynchronously if True
        :param if_exists:
        :param hints:
        :return: None
        """
        return self.parent.delete(self, async_=async_, if_exists=if_exists, hints=hints)

    @utils.with_wait_argument
    def set_storage_tier(self, storage_tier, partition_spec=None, async_=False, hints=None):
        """
        Set storage tier of
        """
        from .tasks import SQLTask

        partition_expr = ""
        if partition_spec is not None:
            if not isinstance(partition_spec, (list, tuple)):
                partition_spec = [partition_spec]
            partition_expr = " " + ", ".join(
                "PARTITION (%s)" % spec for spec in partition_spec
            )

        # as data of partition changed, remove existing download id to avoid TableModified error
        for part in (partition_spec or [None]):
            if isinstance(part, six.string_types):
                part = odps_types.PartitionSpec(part)
            self._download_ids.pop(part, None)

        if isinstance(storage_tier, six.string_types):
            storage_tier = StorageTier(utils.underline_to_camel(storage_tier).lower())

        property_item = "TBLPROPERTIES" if not partition_spec else "PARTITIONPROPERTIES"
        task = SQLTask(
            name='SQLSetStorageTierTask',
            query="ALTER TABLE %s%s SET %s('storagetier'='%s')" % (
                self.full_table_name, partition_expr, property_item, storage_tier.value
            )
        )

        hints = hints or {}
        hints['odps.sql.submit.mode'] = ''
        hints['odps.tiered.storage.enable'] = 'true'
        schema_name = self._get_schema_name()
        if schema_name is not None:
            hints["odps.sql.allow.namespace.schema"] = "true"
            hints["odps.namespace.schema"] = "true"
        task.update_sql_settings(hints)

        instance = self.project.parent[self._client.project].instances.create(task=task)
        self._is_extend_info_loaded = False

        if not async_:
            instance.wait_for_success()
        else:
            return instance

    def new_record(self, values=None):
        """
        Generate a record of the table.

        :param values: the values of this records
        :type values: list
        :return: record
        :rtype: :class:`odps.models.Record`

        :Example:

        >>> table = odps.create_table('test_table', schema=TableSchema.from_lists(['name', 'id'], ['sring', 'string']))
        >>> record = table.new_record()
        >>> record[0] = 'my_name'
        >>> record[1] = 'my_id'
        >>> record = table.new_record(['my_name', 'my_id'])

        .. seealso:: :class:`odps.models.Record`
        """
        return Record(schema=self.table_schema, values=values)

    def to_df(self):
        """
        Create a PyODPS DataFrame from this table.

        :return: DataFrame object
        """
        from ..df import DataFrame

        return DataFrame(self)



# This file is autogenerated by the update_windows_mapping.py script
# Do not edit.
win_tz = {'AUS Central Standard Time': 'Australia/Darwin',
 'AUS Eastern Standard Time': 'Australia/Sydney',
 'Afghanistan Standard Time': 'Asia/Kabul',
 'Alaskan Standard Time': 'America/Anchorage',
 'Aleutian Standard Time': 'America/Adak',
 'Altai Standard Time': 'Asia/Barnaul',
 'Arab Standard Time': 'Asia/Riyadh',
 'Arabian Standard Time': 'Asia/Dubai',
 'Arabic Standard Time': 'Asia/Baghdad',
 'Argentina Standard Time': 'America/Buenos_Aires',
 'Astrakhan Standard Time': 'Europe/Astrakhan',
 'Atlantic Standard Time': 'America/Halifax',
 'Aus Central W. Standard Time': 'Australia/Eucla',
 'Azerbaijan Standard Time': 'Asia/Baku',
 'Azores Standard Time': 'Atlantic/Azores',
 'Bahia Standard Time': 'America/Bahia',
 'Bangladesh Standard Time': 'Asia/Dhaka',
 'Belarus Standard Time': 'Europe/Minsk',
 'Bougainville Standard Time': 'Pacific/Bougainville',
 'Canada Central Standard Time': 'America/Regina',
 'Cape Verde Standard Time': 'Atlantic/Cape_Verde',
 'Caucasus Standard Time': 'Asia/Yerevan',
 'Cen. Australia Standard Time': 'Australia/Adelaide',
 'Central America Standard Time': 'America/Guatemala',
 'Central Asia Standard Time': 'Asia/Almaty',
 'Central Brazilian Standard Time': 'America/Cuiaba',
 'Central Europe Standard Time': 'Europe/Budapest',
 'Central European Standard Time': 'Europe/Warsaw',
 'Central Pacific Standard Time': 'Pacific/Guadalcanal',
 'Central Standard Time': 'America/Chicago',
 'Central Standard Time (Mexico)': 'America/Mexico_City',
 'Chatham Islands Standard Time': 'Pacific/Chatham',
 'China Standard Time': 'Asia/Shanghai',
 'Cuba Standard Time': 'America/Havana',
 'Dateline Standard Time': 'Etc/GMT+12',
 'E. Africa Standard Time': 'Africa/Nairobi',
 'E. Australia Standard Time': 'Australia/Brisbane',
 'E. Europe Standard Time': 'Europe/Chisinau',
 'E. South America Standard Time': 'America/Sao_Paulo',
 'Easter Island Standard Time': 'Pacific/Easter',
 'Eastern Standard Time': 'America/New_York',
 'Eastern Standard Time (Mexico)': 'America/Cancun',
 'Egypt Standard Time': 'Africa/Cairo',
 'Ekaterinburg Standard Time': 'Asia/Yekaterinburg',
 'FLE Standard Time': 'Europe/Kiev',
 'Fiji Standard Time': 'Pacific/Fiji',
 'GMT Standard Time': 'Europe/London',
 'GTB Standard Time': 'Europe/Bucharest',
 'Georgian Standard Time': 'Asia/Tbilisi',
 'Greenland Standard Time': 'America/Godthab',
 'Greenwich Standard Time': 'Atlantic/Reykjavik',
 'Haiti Standard Time': 'America/Port-au-Prince',
 'Hawaiian Standard Time': 'Pacific/Honolulu',
 'India Standard Time': 'Asia/Calcutta',
 'Iran Standard Time': 'Asia/Tehran',
 'Israel Standard Time': 'Asia/Jerusalem',
 'Jordan Standard Time': 'Asia/Amman',
 'Kaliningrad Standard Time': 'Europe/Kaliningrad',
 'Korea Standard Time': 'Asia/Seoul',
 'Libya Standard Time': 'Africa/Tripoli',
 'Line Islands Standard Time': 'Pacific/Kiritimati',
 'Lord Howe Standard Time': 'Australia/Lord_Howe',
 'Magadan Standard Time': 'Asia/Magadan',
 'Magallanes Standard Time': 'America/Punta_Arenas',
 'Marquesas Standard Time': 'Pacific/Marquesas',
 'Mauritius Standard Time': 'Indian/Mauritius',
 'Middle East Standard Time': 'Asia/Beirut',
 'Montevideo Standard Time': 'America/Montevideo',
 'Morocco Standard Time': 'Africa/Casablanca',
 'Mountain Standard Time': 'America/Denver',
 'Mountain Standard Time (Mexico)': 'America/Chihuahua',
 'Myanmar Standard Time': 'Asia/Rangoon',
 'N. Central Asia Standard Time': 'Asia/Novosibirsk',
 'Namibia Standard Time': 'Africa/Windhoek',
 'Nepal Standard Time': 'Asia/Katmandu',
 'New Zealand Standard Time': 'Pacific/Auckland',
 'Newfoundland Standard Time': 'America/St_Johns',
 'Norfolk Standard Time': 'Pacific/Norfolk',
 'North Asia East Standard Time': 'Asia/Irkutsk',
 'North Asia Standard Time': 'Asia/Krasnoyarsk',
 'North Korea Standard Time': 'Asia/Pyongyang',
 'Omsk Standard Time': 'Asia/Omsk',
 'Pacific SA Standard Time': 'America/Santiago',
 'Pacific Standard Time': 'America/Los_Angeles',
 'Pacific Standard Time (Mexico)': 'America/Tijuana',
 'Pakistan Standard Time': 'Asia/Karachi',
 'Paraguay Standard Time': 'America/Asuncion',
 'Romance Standard Time': 'Europe/Paris',
 'Russia Time Zone 10': 'Asia/Srednekolymsk',
 'Russia Time Zone 11': 'Asia/Kamchatka',
 'Russia Time Zone 3': 'Europe/Samara',
 'Russian Standard Time': 'Europe/Moscow',
 'SA Eastern Standard Time': 'America/Cayenne',
 'SA Pacific Standard Time': 'America/Bogota',
 'SA Western Standard Time': 'America/La_Paz',
 'SE Asia Standard Time': 'Asia/Bangkok',
 'Saint Pierre Standard Time': 'America/Miquelon',
 'Sakhalin Standard Time': 'Asia/Sakhalin',
 'Samoa Standard Time': 'Pacific/Apia',
 'Saratov Standard Time': 'Europe/Saratov',
 'Singapore Standard Time': 'Asia/Singapore',
 'South Africa Standard Time': 'Africa/Johannesburg',
 'Sri Lanka Standard Time': 'Asia/Colombo',
 'Syria Standard Time': 'Asia/Damascus',
 'Taipei Standard Time': 'Asia/Taipei',
 'Tasmania Standard Time': 'Australia/Hobart',
 'Tocantins Standard Time': 'America/Araguaina',
 'Tokyo Standard Time': 'Asia/Tokyo',
 'Tomsk Standard Time': 'Asia/Tomsk',
 'Tonga Standard Time': 'Pacific/Tongatapu',
 'Transbaikal Standard Time': 'Asia/Chita',
 'Turkey Standard Time': 'Europe/Istanbul',
 'Turks And Caicos Standard Time': 'America/Grand_Turk',
 'US Eastern Standard Time': 'America/Indianapolis',
 'US Mountain Standard Time': 'America/Phoenix',
 'UTC': 'Etc/GMT',
 'UTC+12': 'Etc/GMT-12',
 'UTC+13': 'Etc/GMT-13',
 'UTC-02': 'Etc/GMT+2',
 'UTC-08': 'Etc/GMT+8',
 'UTC-09': 'Etc/GMT+9',
 'UTC-11': 'Etc/GMT+11',
 'Ulaanbaatar Standard Time': 'Asia/Ulaanbaatar',
 'Venezuela Standard Time': 'America/Caracas',
 'Vladivostok Standard Time': 'Asia/Vladivostok',
 'W. Australia Standard Time': 'Australia/Perth',
 'W. Central Africa Standard Time': 'Africa/Lagos',
 'W. Europe Standard Time': 'Europe/Berlin',
 'W. Mongolia Standard Time': 'Asia/Hovd',
 'West Asia Standard Time': 'Asia/Tashkent',
 'West Bank Standard Time': 'Asia/Hebron',
 'West Pacific Standard Time': 'Pacific/Port_Moresby',
 'Yakutsk Standard Time': 'Asia/Yakutsk'}

# Old name for the win_tz variable:
tz_names = win_tz

tz_win = {'Africa/Abidjan': 'Greenwich Standard Time',
 'Africa/Accra': 'Greenwich Standard Time',
 'Africa/Addis_Ababa': 'E. Africa Standard Time',
 'Africa/Algiers': 'W. Central Africa Standard Time',
 'Africa/Asmera': 'E. Africa Standard Time',
 'Africa/Bamako': 'Greenwich Standard Time',
 'Africa/Bangui': 'W. Central Africa Standard Time',
 'Africa/Banjul': 'Greenwich Standard Time',
 'Africa/Bissau': 'Greenwich Standard Time',
 'Africa/Blantyre': 'South Africa Standard Time',
 'Africa/Brazzaville': 'W. Central Africa Standard Time',
 'Africa/Bujumbura': 'South Africa Standard Time',
 'Africa/Cairo': 'Egypt Standard Time',
 'Africa/Casablanca': 'Morocco Standard Time',
 'Africa/Ceuta': 'Romance Standard Time',
 'Africa/Conakry': 'Greenwich Standard Time',
 'Africa/Dakar': 'Greenwich Standard Time',
 'Africa/Dar_es_Salaam': 'E. Africa Standard Time',
 'Africa/Djibouti': 'E. Africa Standard Time',
 'Africa/Douala': 'W. Central Africa Standard Time',
 'Africa/El_Aaiun': 'Morocco Standard Time',
 'Africa/Freetown': 'Greenwich Standard Time',
 'Africa/Gaborone': 'South Africa Standard Time',
 'Africa/Harare': 'South Africa Standard Time',
 'Africa/Johannesburg': 'South Africa Standard Time',
 'Africa/Juba': 'E. Africa Standard Time',
 'Africa/Kampala': 'E. Africa Standard Time',
 'Africa/Khartoum': 'E. Africa Standard Time',
 'Africa/Kigali': 'South Africa Standard Time',
 'Africa/Kinshasa': 'W. Central Africa Standard Time',
 'Africa/Lagos': 'W. Central Africa Standard Time',
 'Africa/Libreville': 'W. Central Africa Standard Time',
 'Africa/Lome': 'Greenwich Standard Time',
 'Africa/Luanda': 'W. Central Africa Standard Time',
 'Africa/Lubumbashi': 'South Africa Standard Time',
 'Africa/Lusaka': 'South Africa Standard Time',
 'Africa/Malabo': 'W. Central Africa Standard Time',
 'Africa/Maputo': 'South Africa Standard Time',
 'Africa/Maseru': 'South Africa Standard Time',
 'Africa/Mbabane': 'South Africa Standard Time',
 'Africa/Mogadishu': 'E. Africa Standard Time',
 'Africa/Monrovia': 'Greenwich Standard Time',
 'Africa/Nairobi': 'E. Africa Standard Time',
 'Africa/Ndjamena': 'W. Central Africa Standard Time',
 'Africa/Niamey': 'W. Central Africa Standard Time',
 'Africa/Nouakchott': 'Greenwich Standard Time',
 'Africa/Ouagadougou': 'Greenwich Standard Time',
 'Africa/Porto-Novo': 'W. Central Africa Standard Time',
 'Africa/Sao_Tome': 'Greenwich Standard Time',
 'Africa/Timbuktu': 'Greenwich Standard Time',
 'Africa/Tripoli': 'Libya Standard Time',
 'Africa/Tunis': 'W. Central Africa Standard Time',
 'Africa/Windhoek': 'Namibia Standard Time',
 'America/Adak': 'Aleutian Standard Time',
 'America/Anchorage': 'Alaskan Standard Time',
 'America/Anguilla': 'SA Western Standard Time',
 'America/Antigua': 'SA Western Standard Time',
 'America/Araguaina': 'Tocantins Standard Time',
 'America/Argentina/La_Rioja': 'Argentina Standard Time',
 'America/Argentina/Rio_Gallegos': 'Argentina Standard Time',
 'America/Argentina/Salta': 'Argentina Standard Time',
 'America/Argentina/San_Juan': 'Argentina Standard Time',
 'America/Argentina/San_Luis': 'Argentina Standard Time',
 'America/Argentina/Tucuman': 'Argentina Standard Time',
 'America/Argentina/Ushuaia': 'Argentina Standard Time',
 'America/Aruba': 'SA Western Standard Time',
 'America/Asuncion': 'Paraguay Standard Time',
 'America/Atka': 'Aleutian Standard Time',
 'America/Bahia': 'Bahia Standard Time',
 'America/Bahia_Banderas': 'Central Standard Time (Mexico)',
 'America/Barbados': 'SA Western Standard Time',
 'America/Belem': 'SA Eastern Standard Time',
 'America/Belize': 'Central America Standard Time',
 'America/Blanc-Sablon': 'SA Western Standard Time',
 'America/Boa_Vista': 'SA Western Standard Time',
 'America/Bogota': 'SA Pacific Standard Time',
 'America/Boise': 'Mountain Standard Time',
 'America/Buenos_Aires': 'Argentina Standard Time',
 'America/Cambridge_Bay': 'Mountain Standard Time',
 'America/Campo_Grande': 'Central Brazilian Standard Time',
 'America/Cancun': 'Eastern Standard Time (Mexico)',
 'America/Caracas': 'Venezuela Standard Time',
 'America/Catamarca': 'Argentina Standard Time',
 'America/Cayenne': 'SA Eastern Standard Time',
 'America/Cayman': 'SA Pacific Standard Time',
 'America/Chicago': 'Central Standard Time',
 'America/Chihuahua': 'Mountain Standard Time (Mexico)',
 'America/Coral_Harbour': 'SA Pacific Standard Time',
 'America/Cordoba': 'Argentina Standard Time',
 'America/Costa_Rica': 'Central America Standard Time',
 'America/Creston': 'US Mountain Standard Time',
 'America/Cuiaba': 'Central Brazilian Standard Time',
 'America/Curacao': 'SA Western Standard Time',
 'America/Danmarkshavn': 'UTC',
 'America/Dawson': 'Pacific Standard Time',
 'America/Dawson_Creek': 'US Mountain Standard Time',
 'America/Denver': 'Mountain Standard Time',
 'America/Detroit': 'Eastern Standard Time',
 'America/Dominica': 'SA Western Standard Time',
 'America/Edmonton': 'Mountain Standard Time',
 'America/Eirunepe': 'SA Pacific Standard Time',
 'America/El_Salvador': 'Central America Standard Time',
 'America/Ensenada': 'Pacific Standard Time (Mexico)',
 'America/Fort_Nelson': 'US Mountain Standard Time',
 'America/Fortaleza': 'SA Eastern Standard Time',
 'America/Glace_Bay': 'Atlantic Standard Time',
 'America/Godthab': 'Greenland Standard Time',
 'America/Goose_Bay': 'Atlantic Standard Time',
 'America/Grand_Turk': 'Turks And Caicos Standard Time',
 'America/Grenada': 'SA Western Standard Time',
 'America/Guadeloupe': 'SA Western Standard Time',
 'America/Guatemala': 'Central America Standard Time',
 'America/Guayaquil': 'SA Pacific Standard Time',
 'America/Guyana': 'SA Western Standard Time',
 'America/Halifax': 'Atlantic Standard Time',
 'America/Havana': 'Cuba Standard Time',
 'America/Hermosillo': 'US Mountain Standard Time',
 'America/Indiana/Knox': 'Central Standard Time',
 'America/Indiana/Marengo': 'US Eastern Standard Time',
 'America/Indiana/Petersburg': 'Eastern Standard Time',
 'America/Indiana/Tell_City': 'Central Standard Time',
 'America/Indiana/Vevay': 'US Eastern Standard Time',
 'America/Indiana/Vincennes': 'Eastern Standard Time',
 'America/Indiana/Winamac': 'Eastern Standard Time',
 'America/Indianapolis': 'US Eastern Standard Time',
 'America/Inuvik': 'Mountain Standard Time',
 'America/Iqaluit': 'Eastern Standard Time',
 'America/Jamaica': 'SA Pacific Standard Time',
 'America/Jujuy': 'Argentina Standard Time',
 'America/Juneau': 'Alaskan Standard Time',
 'America/Kentucky/Monticello': 'Eastern Standard Time',
 'America/Knox_IN': 'Central Standard Time',
 'America/Kralendijk': 'SA Western Standard Time',
 'America/La_Paz': 'SA Western Standard Time',
 'America/Lima': 'SA Pacific Standard Time',
 'America/Los_Angeles': 'Pacific Standard Time',
 'America/Louisville': 'Eastern Standard Time',
 'America/Lower_Princes': 'SA Western Standard Time',
 'America/Maceio': 'SA Eastern Standard Time',
 'America/Managua': 'Central America Standard Time',
 'America/Manaus': 'SA Western Standard Time',
 'America/Marigot': 'SA Western Standard Time',
 'America/Martinique': 'SA Western Standard Time',
 'America/Matamoros': 'Central Standard Time',
 'America/Mazatlan': 'Mountain Standard Time (Mexico)',
 'America/Mendoza': 'Argentina Standard Time',
 'America/Menominee': 'Central Standard Time',
 'America/Merida': 'Central Standard Time (Mexico)',
 'America/Metlakatla': 'Alaskan Standard Time',
 'America/Mexico_City': 'Central Standard Time (Mexico)',
 'America/Miquelon': 'Saint Pierre Standard Time',
 'America/Moncton': 'Atlantic Standard Time',
 'America/Monterrey': 'Central Standard Time (Mexico)',
 'America/Montevideo': 'Montevideo Standard Time',
 'America/Montreal': 'Eastern Standard Time',
 'America/Montserrat': 'SA Western Standard Time',
 'America/Nassau': 'Eastern Standard Time',
 'America/New_York': 'Eastern Standard Time',
 'America/Nipigon': 'Eastern Standard Time',
 'America/Nome': 'Alaskan Standard Time',
 'America/Noronha': 'UTC-02',
 'America/North_Dakota/Beulah': 'Central Standard Time',
 'America/North_Dakota/Center': 'Central Standard Time',
 'America/North_Dakota/New_Salem': 'Central Standard Time',
 'America/Ojinaga': 'Mountain Standard Time',
 'America/Panama': 'SA Pacific Standard Time',
 'America/Pangnirtung': 'Eastern Standard Time',
 'America/Paramaribo': 'SA Eastern Standard Time',
 'America/Phoenix': 'US Mountain Standard Time',
 'America/Port-au-Prince': 'Haiti Standard Time',
 'America/Port_of_Spain': 'SA Western Standard Time',
 'America/Porto_Acre': 'SA Pacific Standard Time',
 'America/Porto_Velho': 'SA Western Standard Time',
 'America/Puerto_Rico': 'SA Western Standard Time',
 'America/Punta_Arenas': 'Magallanes Standard Time',
 'America/Rainy_River': 'Central Standard Time',
 'America/Rankin_Inlet': 'Central Standard Time',
 'America/Recife': 'SA Eastern Standard Time',
 'America/Regina': 'Canada Central Standard Time',
 'America/Resolute': 'Central Standard Time',
 'America/Rio_Branco': 'SA Pacific Standard Time',
 'America/Santa_Isabel': 'Pacific Standard Time (Mexico)',
 'America/Santarem': 'SA Eastern Standard Time',
 'America/Santiago': 'Pacific SA Standard Time',
 'America/Santo_Domingo': 'SA Western Standard Time',
 'America/Sao_Paulo': 'E. South America Standard Time',
 'America/Scoresbysund': 'Azores Standard Time',
 'America/Shiprock': 'Mountain Standard Time',
 'America/Sitka': 'Alaskan Standard Time',
 'America/St_Barthelemy': 'SA Western Standard Time',
 'America/St_Johns': 'Newfoundland Standard Time',
 'America/St_Kitts': 'SA Western Standard Time',
 'America/St_Lucia': 'SA Western Standard Time',
 'America/St_Thomas': 'SA Western Standard Time',
 'America/St_Vincent': 'SA Western Standard Time',
 'America/Swift_Current': 'Canada Central Standard Time',
 'America/Tegucigalpa': 'Central America Standard Time',
 'America/Thule': 'Atlantic Standard Time',
 'America/Thunder_Bay': 'Eastern Standard Time',
 'America/Tijuana': 'Pacific Standard Time (Mexico)',
 'America/Toronto': 'Eastern Standard Time',
 'America/Tortola': 'SA Western Standard Time',
 'America/Vancouver': 'Pacific Standard Time',
 'America/Virgin': 'SA Western Standard Time',
 'America/Whitehorse': 'Pacific Standard Time',
 'America/Winnipeg': 'Central Standard Time',
 'America/Yakutat': 'Alaskan Standard Time',
 'America/Yellowknife': 'Mountain Standard Time',
 'Antarctica/Casey': 'Central Pacific Standard Time',
 'Antarctica/Davis': 'SE Asia Standard Time',
 'Antarctica/DumontDUrville': 'West Pacific Standard Time',
 'Antarctica/Macquarie': 'Central Pacific Standard Time',
 'Antarctica/Mawson': 'West Asia Standard Time',
 'Antarctica/McMurdo': 'New Zealand Standard Time',
 'Antarctica/Palmer': 'Magallanes Standard Time',
 'Antarctica/Rothera': 'SA Eastern Standard Time',
 'Antarctica/South_Pole': 'New Zealand Standard Time',
 'Antarctica/Syowa': 'E. Africa Standard Time',
 'Antarctica/Vostok': 'Central Asia Standard Time',
 'Arctic/Longyearbyen': 'W. Europe Standard Time',
 'Asia/Aden': 'Arab Standard Time',
 'Asia/Almaty': 'Central Asia Standard Time',
 'Asia/Amman': 'Jordan Standard Time',
 'Asia/Anadyr': 'Russia Time Zone 11',
 'Asia/Aqtau': 'West Asia Standard Time',
 'Asia/Aqtobe': 'West Asia Standard Time',
 'Asia/Ashgabat': 'West Asia Standard Time',
 'Asia/Ashkhabad': 'West Asia Standard Time',
 'Asia/Atyrau': 'West Asia Standard Time',
 'Asia/Baghdad': 'Arabic Standard Time',
 'Asia/Bahrain': 'Arab Standard Time',
 'Asia/Baku': 'Azerbaijan Standard Time',
 'Asia/Bangkok': 'SE Asia Standard Time',
 'Asia/Barnaul': 'Altai Standard Time',
 'Asia/Beirut': 'Middle East Standard Time',
 'Asia/Bishkek': 'Central Asia Standard Time',
 'Asia/Brunei': 'Singapore Standard Time',
 'Asia/Calcutta': 'India Standard Time',
 'Asia/Chita': 'Transbaikal Standard Time',
 'Asia/Choibalsan': 'Ulaanbaatar Standard Time',
 'Asia/Chongqing': 'China Standard Time',
 'Asia/Chungking': 'China Standard Time',
 'Asia/Colombo': 'Sri Lanka Standard Time',
 'Asia/Dacca': 'Bangladesh Standard Time',
 'Asia/Damascus': 'Syria Standard Time',
 'Asia/Dhaka': 'Bangladesh Standard Time',
 'Asia/Dili': 'Tokyo Standard Time',
 'Asia/Dubai': 'Arabian Standard Time',
 'Asia/Dushanbe': 'West Asia Standard Time',
 'Asia/Famagusta': 'Turkey Standard Time',
 'Asia/Gaza': 'West Bank Standard Time',
 'Asia/Harbin': 'China Standard Time',
 'Asia/Hebron': 'West Bank Standard Time',
 'Asia/Hong_Kong': 'China Standard Time',
 'Asia/Hovd': 'W. Mongolia Standard Time',
 'Asia/Irkutsk': 'North Asia East Standard Time',
 'Asia/Jakarta': 'SE Asia Standard Time',
 'Asia/Jayapura': 'Tokyo Standard Time',
 'Asia/Jerusalem': 'Israel Standard Time',
 'Asia/Kabul': 'Afghanistan Standard Time',
 'Asia/Kamchatka': 'Russia Time Zone 11',
 'Asia/Karachi': 'Pakistan Standard Time',
 'Asia/Kashgar': 'Central Asia Standard Time',
 'Asia/Katmandu': 'Nepal Standard Time',
 'Asia/Khandyga': 'Yakutsk Standard Time',
 'Asia/Krasnoyarsk': 'North Asia Standard Time',
 'Asia/Kuala_Lumpur': 'Singapore Standard Time',
 'Asia/Kuching': 'Singapore Standard Time',
 'Asia/Kuwait': 'Arab Standard Time',
 'Asia/Macao': 'China Standard Time',
 'Asia/Macau': 'China Standard Time',
 'Asia/Magadan': 'Magadan Standard Time',
 'Asia/Makassar': 'Singapore Standard Time',
 'Asia/Manila': 'Singapore Standard Time',
 'Asia/Muscat': 'Arabian Standard Time',
 'Asia/Nicosia': 'GTB Standard Time',
 'Asia/Novokuznetsk': 'North Asia Standard Time',
 'Asia/Novosibirsk': 'N. Central Asia Standard Time',
 'Asia/Omsk': 'Omsk Standard Time',
 'Asia/Oral': 'West Asia Standard Time',
 'Asia/Phnom_Penh': 'SE Asia Standard Time',
 'Asia/Pontianak': 'SE Asia Standard Time',
 'Asia/Pyongyang': 'North Korea Standard Time',
 'Asia/Qatar': 'Arab Standard Time',
 'Asia/Qyzylorda': 'Central Asia Standard Time',
 'Asia/Rangoon': 'Myanmar Standard Time',
 'Asia/Riyadh': 'Arab Standard Time',
 'Asia/Saigon': 'SE Asia Standard Time',
 'Asia/Sakhalin': 'Sakhalin Standard Time',
 'Asia/Samarkand': 'West Asia Standard Time',
 'Asia/Seoul': 'Korea Standard Time',
 'Asia/Shanghai': 'China Standard Time',
 'Asia/Singapore': 'Singapore Standard Time',
 'Asia/Srednekolymsk': 'Russia Time Zone 10',
 'Asia/Taipei': 'Taipei Standard Time',
 'Asia/Tashkent': 'West Asia Standard Time',
 'Asia/Tbilisi': 'Georgian Standard Time',
 'Asia/Tehran': 'Iran Standard Time',
 'Asia/Tel_Aviv': 'Israel Standard Time',
 'Asia/Thimbu': 'Bangladesh Standard Time',
 'Asia/Thimphu': 'Bangladesh Standard Time',
 'Asia/Tokyo': 'Tokyo Standard Time',
 'Asia/Tomsk': 'Tomsk Standard Time',
 'Asia/Ujung_Pandang': 'Singapore Standard Time',
 'Asia/Ulaanbaatar': 'Ulaanbaatar Standard Time',
 'Asia/Ulan_Bator': 'Ulaanbaatar Standard Time',
 'Asia/Urumqi': 'Central Asia Standard Time',
 'Asia/Ust-Nera': 'Vladivostok Standard Time',
 'Asia/Vientiane': 'SE Asia Standard Time',
 'Asia/Vladivostok': 'Vladivostok Standard Time',
 'Asia/Yakutsk': 'Yakutsk Standard Time',
 'Asia/Yekaterinburg': 'Ekaterinburg Standard Time',
 'Asia/Yerevan': 'Caucasus Standard Time',
 'Atlantic/Azores': 'Azores Standard Time',
 'Atlantic/Bermuda': 'Atlantic Standard Time',
 'Atlantic/Canary': 'GMT Standard Time',
 'Atlantic/Cape_Verde': 'Cape Verde Standard Time',
 'Atlantic/Faeroe': 'GMT Standard Time',
 'Atlantic/Jan_Mayen': 'W. Europe Standard Time',
 'Atlantic/Madeira': 'GMT Standard Time',
 'Atlantic/Reykjavik': 'Greenwich Standard Time',
 'Atlantic/South_Georgia': 'UTC-02',
 'Atlantic/St_Helena': 'Greenwich Standard Time',
 'Atlantic/Stanley': 'SA Eastern Standard Time',
 'Australia/ACT': 'AUS Eastern Standard Time',
 'Australia/Adelaide': 'Cen. Australia Standard Time',
 'Australia/Brisbane': 'E. Australia Standard Time',
 'Australia/Broken_Hill': 'Cen. Australia Standard Time',
 'Australia/Canberra': 'AUS Eastern Standard Time',
 'Australia/Currie': 'Tasmania Standard Time',
 'Australia/Darwin': 'AUS Central Standard Time',
 'Australia/Eucla': 'Aus Central W. Standard Time',
 'Australia/Hobart': 'Tasmania Standard Time',
 'Australia/LHI': 'Lord Howe Standard Time',
 'Australia/Lindeman': 'E. Australia Standard Time',
 'Australia/Lord_Howe': 'Lord Howe Standard Time',
 'Australia/Melbourne': 'AUS Eastern Standard Time',
 'Australia/NSW': 'AUS Eastern Standard Time',
 'Australia/North': 'AUS Central Standard Time',
 'Australia/Perth': 'W. Australia Standard Time',
 'Australia/Queensland': 'E. Australia Standard Time',
 'Australia/South': 'Cen. Australia Standard Time',
 'Australia/Sydney': 'AUS Eastern Standard Time',
 'Australia/Tasmania': 'Tasmania Standard Time',
 'Australia/Victoria': 'AUS Eastern Standard Time',
 'Australia/West': 'W. Australia Standard Time',
 'Australia/Yancowinna': 'Cen. Australia Standard Time',
 'Brazil/Acre': 'SA Pacific Standard Time',
 'Brazil/DeNoronha': 'UTC-02',
 'Brazil/East': 'E. South America Standard Time',
 'Brazil/West': 'SA Western Standard Time',
 'CST6CDT': 'Central Standard Time',
 'Canada/Atlantic': 'Atlantic Standard Time',
 'Canada/Central': 'Central Standard Time',
 'Canada/Eastern': 'Eastern Standard Time',
 'Canada/Mountain': 'Mountain Standard Time',
 'Canada/Newfoundland': 'Newfoundland Standard Time',
 'Canada/Pacific': 'Pacific Standard Time',
 'Canada/Saskatchewan': 'Canada Central Standard Time',
 'Canada/Yukon': 'Pacific Standard Time',
 'Chile/Continental': 'Pacific SA Standard Time',
 'Chile/EasterIsland': 'Easter Island Standard Time',
 'Cuba': 'Cuba Standard Time',
 'EST5EDT': 'Eastern Standard Time',
 'Egypt': 'Egypt Standard Time',
 'Eire': 'GMT Standard Time',
 'Etc/GMT': 'UTC',
 'Etc/GMT+1': 'Cape Verde Standard Time',
 'Etc/GMT+10': 'Hawaiian Standard Time',
 'Etc/GMT+11': 'UTC-11',
 'Etc/GMT+12': 'Dateline Standard Time',
 'Etc/GMT+2': 'UTC-02',
 'Etc/GMT+3': 'SA Eastern Standard Time',
 'Etc/GMT+4': 'SA Western Standard Time',
 'Etc/GMT+5': 'SA Pacific Standard Time',
 'Etc/GMT+6': 'Central America Standard Time',
 'Etc/GMT+7': 'US Mountain Standard Time',
 'Etc/GMT+8': 'UTC-08',
 'Etc/GMT+9': 'UTC-09',
 'Etc/GMT-1': 'W. Central Africa Standard Time',
 'Etc/GMT-10': 'West Pacific Standard Time',
 'Etc/GMT-11': 'Central Pacific Standard Time',
 'Etc/GMT-12': 'UTC+12',
 'Etc/GMT-13': 'UTC+13',
 'Etc/GMT-14': 'Line Islands Standard Time',
 'Etc/GMT-2': 'South Africa Standard Time',
 'Etc/GMT-3': 'E. Africa Standard Time',
 'Etc/GMT-4': 'Arabian Standard Time',
 'Etc/GMT-5': 'West Asia Standard Time',
 'Etc/GMT-6': 'Central Asia Standard Time',
 'Etc/GMT-7': 'SE Asia Standard Time',
 'Etc/GMT-8': 'Singapore Standard Time',
 'Etc/GMT-9': 'Tokyo Standard Time',
 'Etc/UTC': 'UTC',
 'Europe/Amsterdam': 'W. Europe Standard Time',
 'Europe/Andorra': 'W. Europe Standard Time',
 'Europe/Astrakhan': 'Astrakhan Standard Time',
 'Europe/Athens': 'GTB Standard Time',
 'Europe/Belfast': 'GMT Standard Time',
 'Europe/Belgrade': 'Central Europe Standard Time',
 'Europe/Berlin': 'W. Europe Standard Time',
 'Europe/Bratislava': 'Central Europe Standard Time',
 'Europe/Brussels': 'Romance Standard Time',
 'Europe/Bucharest': 'GTB Standard Time',
 'Europe/Budapest': 'Central Europe Standard Time',
 'Europe/Busingen': 'W. Europe Standard Time',
 'Europe/Chisinau': 'E. Europe Standard Time',
 'Europe/Copenhagen': 'Romance Standard Time',
 'Europe/Dublin': 'GMT Standard Time',
 'Europe/Gibraltar': 'W. Europe Standard Time',
 'Europe/Guernsey': 'GMT Standard Time',
 'Europe/Helsinki': 'FLE Standard Time',
 'Europe/Isle_of_Man': 'GMT Standard Time',
 'Europe/Istanbul': 'Turkey Standard Time',
 'Europe/Jersey': 'GMT Standard Time',
 'Europe/Kaliningrad': 'Kaliningrad Standard Time',
 'Europe/Kiev': 'FLE Standard Time',
 'Europe/Kirov': 'Russian Standard Time',
 'Europe/Lisbon': 'GMT Standard Time',
 'Europe/Ljubljana': 'Central Europe Standard Time',
 'Europe/London': 'GMT Standard Time',
 'Europe/Luxembourg': 'W. Europe Standard Time',
 'Europe/Madrid': 'Romance Standard Time',
 'Europe/Malta': 'W. Europe Standard Time',
 'Europe/Mariehamn': 'FLE Standard Time',
 'Europe/Minsk': 'Belarus Standard Time',
 'Europe/Monaco': 'W. Europe Standard Time',
 'Europe/Moscow': 'Russian Standard Time',
 'Europe/Oslo': 'W. Europe Standard Time',
 'Europe/Paris': 'Romance Standard Time',
 'Europe/Podgorica': 'Central Europe Standard Time',
 'Europe/Prague': 'Central Europe Standard Time',
 'Europe/Riga': 'FLE Standard Time',
 'Europe/Rome': 'W. Europe Standard Time',
 'Europe/Samara': 'Russia Time Zone 3',
 'Europe/San_Marino': 'W. Europe Standard Time',
 'Europe/Sarajevo': 'Central European Standard Time',
 'Europe/Saratov': 'Saratov Standard Time',
 'Europe/Simferopol': 'Russian Standard Time',
 'Europe/Skopje': 'Central European Standard Time',
 'Europe/Sofia': 'FLE Standard Time',
 'Europe/Stockholm': 'W. Europe Standard Time',
 'Europe/Tallinn': 'FLE Standard Time',
 'Europe/Tirane': 'Central Europe Standard Time',
 'Europe/Tiraspol': 'E. Europe Standard Time',
 'Europe/Ulyanovsk': 'Astrakhan Standard Time',
 'Europe/Uzhgorod': 'FLE Standard Time',
 'Europe/Vaduz': 'W. Europe Standard Time',
 'Europe/Vatican': 'W. Europe Standard Time',
 'Europe/Vienna': 'W. Europe Standard Time',
 'Europe/Vilnius': 'FLE Standard Time',
 'Europe/Volgograd': 'Russian Standard Time',
 'Europe/Warsaw': 'Central European Standard Time',
 'Europe/Zagreb': 'Central European Standard Time',
 'Europe/Zaporozhye': 'FLE Standard Time',
 'Europe/Zurich': 'W. Europe Standard Time',
 'GB': 'GMT Standard Time',
 'GB-Eire': 'GMT Standard Time',
 'GMT+0': 'UTC',
 'GMT-0': 'UTC',
 'GMT0': 'UTC',
 'Greenwich': 'UTC',
 'Hongkong': 'China Standard Time',
 'Iceland': 'Greenwich Standard Time',
 'Indian/Antananarivo': 'E. Africa Standard Time',
 'Indian/Chagos': 'Central Asia Standard Time',
 'Indian/Christmas': 'SE Asia Standard Time',
 'Indian/Cocos': 'Myanmar Standard Time',
 'Indian/Comoro': 'E. Africa Standard Time',
 'Indian/Kerguelen': 'West Asia Standard Time',
 'Indian/Mahe': 'Mauritius Standard Time',
 'Indian/Maldives': 'West Asia Standard Time',
 'Indian/Mauritius': 'Mauritius Standard Time',
 'Indian/Mayotte': 'E. Africa Standard Time',
 'Indian/Reunion': 'Mauritius Standard Time',
 'Iran': 'Iran Standard Time',
 'Israel': 'Israel Standard Time',
 'Jamaica': 'SA Pacific Standard Time',
 'Japan': 'Tokyo Standard Time',
 'Kwajalein': 'UTC+12',
 'Libya': 'Libya Standard Time',
 'MST7MDT': 'Mountain Standard Time',
 'Mexico/BajaNorte': 'Pacific Standard Time (Mexico)',
 'Mexico/BajaSur': 'Mountain Standard Time (Mexico)',
 'Mexico/General': 'Central Standard Time (Mexico)',
 'NZ': 'New Zealand Standard Time',
 'NZ-CHAT': 'Chatham Islands Standard Time',
 'Navajo': 'Mountain Standard Time',
 'PRC': 'China Standard Time',
 'PST8PDT': 'Pacific Standard Time',
 'Pacific/Apia': 'Samoa Standard Time',
 'Pacific/Auckland': 'New Zealand Standard Time',
 'Pacific/Bougainville': 'Bougainville Standard Time',
 'Pacific/Chatham': 'Chatham Islands Standard Time',
 'Pacific/Easter': 'Easter Island Standard Time',
 'Pacific/Efate': 'Central Pacific Standard Time',
 'Pacific/Enderbury': 'UTC+13',
 'Pacific/Fakaofo': 'UTC+13',
 'Pacific/Fiji': 'Fiji Standard Time',
 'Pacific/Funafuti': 'UTC+12',
 'Pacific/Galapagos': 'Central America Standard Time',
 'Pacific/Gambier': 'UTC-09',
 'Pacific/Guadalcanal': 'Central Pacific Standard Time',
 'Pacific/Guam': 'West Pacific Standard Time',
 'Pacific/Honolulu': 'Hawaiian Standard Time',
 'Pacific/Johnston': 'Hawaiian Standard Time',
 'Pacific/Kiritimati': 'Line Islands Standard Time',
 'Pacific/Kosrae': 'Central Pacific Standard Time',
 'Pacific/Kwajalein': 'UTC+12',
 'Pacific/Majuro': 'UTC+12',
 'Pacific/Marquesas': 'Marquesas Standard Time',
 'Pacific/Midway': 'UTC-11',
 'Pacific/Nauru': 'UTC+12',
 'Pacific/Niue': 'UTC-11',
 'Pacific/Norfolk': 'Norfolk Standard Time',
 'Pacific/Noumea': 'Central Pacific Standard Time',
 'Pacific/Pago_Pago': 'UTC-11',
 'Pacific/Palau': 'Tokyo Standard Time',
 'Pacific/Pitcairn': 'UTC-08',
 'Pacific/Ponape': 'Central Pacific Standard Time',
 'Pacific/Port_Moresby': 'West Pacific Standard Time',
 'Pacific/Rarotonga': 'Hawaiian Standard Time',
 'Pacific/Saipan': 'West Pacific Standard Time',
 'Pacific/Samoa': 'UTC-11',
 'Pacific/Tahiti': 'Hawaiian Standard Time',
 'Pacific/Tarawa': 'UTC+12',
 'Pacific/Tongatapu': 'Tonga Standard Time',
 'Pacific/Truk': 'West Pacific Standard Time',
 'Pacific/Wake': 'UTC+12',
 'Pacific/Wallis': 'UTC+12',
 'Poland': 'Central European Standard Time',
 'Portugal': 'GMT Standard Time',
 'ROC': 'Taipei Standard Time',
 'ROK': 'Korea Standard Time',
 'Singapore': 'Singapore Standard Time',
 'Turkey': 'Turkey Standard Time',
 'US/Alaska': 'Alaskan Standard Time',
 'US/Aleutian': 'Aleutian Standard Time',
 'US/Arizona': 'US Mountain Standard Time',
 'US/Central': 'Central Standard Time',
 'US/Eastern': 'Eastern Standard Time',
 'US/Hawaii': 'Hawaiian Standard Time',
 'US/Indiana-Starke': 'Central Standard Time',
 'US/Michigan': 'Eastern Standard Time',
 'US/Mountain': 'Mountain Standard Time',
 'US/Pacific': 'Pacific Standard Time',
 'US/Samoa': 'UTC-11',
 'UTC': 'UTC',
 'Universal': 'UTC',
 'W-SU': 'Russian Standard Time',
 'Zulu': 'UTC'}



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ..enums import FieldRole
from .utils import get_field_name_by_role, metrics_result

logger = logging.getLogger(__name__)


def _run_evaluation_node(df, col_true, col_pred, execute_now=True, result_callback=None):
    from . import _customize
    eval_fun = getattr(_customize, '_eval_regression')
    return eval_fun(df, label_col=col_true, predict_col=col_pred,
                    execute_now=execute_now, _result_callback=result_callback)


@metrics_result(_run_evaluation_node)
def mean_squared_error(df, col_true, col_pred=None):
    """
    Compute mean squared error of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean squared error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['mse']


@metrics_result(_run_evaluation_node)
def mean_absolute_error(df, col_true, col_pred=None):
    """
    Compute mean absolute error of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['mae']


@metrics_result(_run_evaluation_node)
def mean_absolute_percentage_error(df, col_true, col_pred=None):
    """
    Compute mean absolute percentage error of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute percentage error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['mape']


@metrics_result(_run_evaluation_node)
def total_sum_of_squares(df, col_true, col_pred=None):
    """
    Compute total sum of squares (SST) of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute percentage error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['sst']


@metrics_result(_run_evaluation_node)
def explained_sum_of_squares(df, col_true, col_pred=None):
    """
    Compute explained sum of squares (SSE) of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute percentage error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['sse']


@metrics_result(_run_evaluation_node)
def r2_score(df, col_true, col_pred=None):
    """
    Compute determination coefficient (R2) of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute percentage error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['r2']


@metrics_result(_run_evaluation_node)
def multi_corr(df, col_true, col_pred=None):
    """
    Compute multiple correlation coefficient (R) of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute percentage error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['r']


@metrics_result(_run_evaluation_node)
def rooted_mean_squared_error(df, col_true, col_pred=None):
    """
    Compute rooted mean squared error (RMSE) of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute percentage error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['rmse']


@metrics_result(_run_evaluation_node)
def mean_absolute_deviation(df, col_true, col_pred=None):
    """
    Compute mean absolute deviation (MAD) of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: Mean absolute percentage error
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['mad']


@metrics_result(_run_evaluation_node)
def residual_histogram(df, col_true, col_pred=None):
    """
    Compute histogram of residuals of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true value
    :type col_true: str
    :param col_true: column name of predicted value, 'prediction_score' by default.
    :type col_pred: str
    :return: histograms for every columns, containing histograms and bins.
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_VALUE)
    return _run_evaluation_node(df, col_true, col_pred)['hist']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import Iterable
from .resource import Resource, FileResource
from .. import serializers, errors
from ..compat import six

DEFAULT_RESOURCE_CHUNK_SIZE = 64 << 20


class Resources(Iterable):

    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems')
    resources = serializers.XMLNodesReferencesField(Resource, 'Resource')

    def get_typed(self, name, type, **kw):
        type_cls = Resource._get_cls(type)
        parent = self

        project_name, schema_name, name = Resource.split_resource_name(name)
        if project_name is not None:
            parent = self.parent.project.parent[project_name]
            if schema_name is not None:
                parent = parent.schemas[schema_name]
            parent = parent.resources
        if "temp" in kw:
            kw["is_temp_resource"] = kw.pop("temp")
        return type_cls(client=self._client, parent=parent, name=name, **kw)

    def _get(self, name):
        return self.get_typed(name, None)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            try:
                resource = self._get(item)
            except errors.NoSuchObject:
                return False
        elif isinstance(item, Resource):
            resource = item
        else:
            return False

        try:
            resource.reload()
            return True
        except errors.NoSuchObject:
            return False

    def __iter__(self):
        return self.iterate()

    def get(self, name, type=None):
        return self.get_typed(name, type)

    def iterate(self, name=None, owner=None):
        params = {'expectmarker': 'true'}
        if name is not None:
            params['name'] = name
        if owner is not None:
            params['owner'] = owner
        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['curr_schema'] = schema_name

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, params=params)

            r = Resources.parse(self._client, resp, obj=self)
            params['marker'] = r.marker

            return r.resources

        while True:
            resources = _it()
            if resources is None:
                break
            for resource in resources:
                yield resource

    def create(self, obj=None, **kwargs):
        if obj is None and 'type' not in kwargs:
            raise ValueError('Unknown resource type to create.')

        if 'temp' in kwargs:
            kwargs['is_temp_resource'] = kwargs.pop('temp')
        if 'part' in kwargs:
            kwargs['is_part_resource'] = kwargs.pop('part')

        ctor_kw = kwargs.copy()
        ctor_kw.pop('file_obj', None)
        ctor_kw.pop('fileobj', None)
        obj = obj or Resource(parent=self, client=self._client, **ctor_kw)

        if obj.type == Resource.Type.UNKOWN:
            raise ValueError('Unknown resource type to create.')
        if obj.parent is None:
            obj._parent = self
        if obj._client is None:
            obj._client = self._client
        return obj.create(overwrite=False, **kwargs)

    def update(self, obj, **kwargs):
        return obj.create(overwrite=True, **kwargs)

    def delete(self, name):
        if not isinstance(name, Resource):
            resource = Resource(name=name, parent=self, type=Resource.Type.UNKOWN)
        else:
            resource = name
            name = name.name
        del self[name]  # release cache

        url = resource.resource()
        self._client.delete(url, curr_schema=self._get_schema_name())

    def _request(self, name, stream=False, offset=None, read_size=None):
        if isinstance(name, FileResource):
            res = name
        else:
            res = Resource(name, parent=self, client=self._client)
        url = res.resource()

        headers = {'Content-Type': 'application/octet-stream'}
        params = {}
        if offset is not None:
            params["rOffset"] = str(offset)
        if read_size is not None:
            params["rSize"] = str(read_size)

        resp = self._client.get(
            url, headers=headers, params=params, stream=stream, curr_schema=self._get_schema_name()
        )
        return resp

    def iter_resource_content(self, name, text_mode=False):
        resp = self._request(name, stream=True)

        return resp.iter_content(decode_unicode=text_mode)

    def read_resource(
        self, name, encoding='utf-8', text_mode=False, offset=None, read_size=None
    ):
        resp = self._request(name, offset=offset, read_size=read_size)

        content = resp.content
        if not text_mode:
            if isinstance(content, six.text_type):
                content = content.encode(encoding)  # read as bytes
            sio = six.BytesIO(content)
        else:
            if isinstance(content, six.binary_type):
                content = content.decode(encoding)
            sio = six.StringIO(content)

        has_remaining = resp.headers.get("x-odps-resource-has-remaining") or "false"
        sio.is_eof = has_remaining.lower() != "true"
        return sio

    def merge_part_files(self, resource, part_resources, md5_hex, overwrite=False):
        content = md5_hex + "|" + ",".join(res.name for res in part_resources)
        total_bytes = sum(res.size for res in part_resources)
        resource_args = resource.extract()
        resource_args.update(
            {"parent": self, "client": self._client, "merge_total_bytes": total_bytes}
        )
        merge_res = Resource(**resource_args)
        merge_res.create(overwrite=overwrite, fileobj=content)



#!/usr/bin/env python
from __future__ import print_function
import argparse
import contextlib
import errno
import glob
import json
import logging
import os
import platform
import re
import shlex
import shutil
import subprocess
import sys
import threading
import time
import warnings
from collections import defaultdict

_DEFAULT_PYABI = "cp37-cp37m"
_MCPY27_PYABI = "cp27-cp27m"
_DWPY27_PYABI = "cp27-cp27mu"
_DEFAULT_DOCKER_IMAGE_X64 = "quay.io/pypa/manylinux2014_x86_64"
_DEFAULT_DOCKER_IMAGE_ARM64 = "quay.io/pypa/manylinux2014_aarch64"
_LEGACY_DOCKER_IMAGE_X64 = "quay.io/pypa/manylinux1_x86_64"
_DEFAULT_PACKAGE_SITE = "packages"
_DEFAULT_OUTPUT_FILE = "packages.tar.gz"
_REQUIREMENT_FILE_NAME = "requirements-extra.txt"
_INSTALL_REQ_FILE_NAME = "install-requires.txt"
_EXCLUDE_FILE_NAME = "excludes.txt"
_BEFORE_SCRIPT_FILE_NAME = "before-script.sh"
_VCS_FILE_NAME = "requirements-vcs.txt"
_PACK_SCRIPT_FILE_NAME = "pack.sh"
_DYNLIB_PYPROJECT_TOML_FILE_NAME = "dynlib.toml"
_DEFAULT_MINIKUBE_USER_ROOT = "/pypack_user_home"

_SEGFAULT_ERR_CODE = 139

python_abi_env = os.getenv("PYABI")
cmd_before_build = os.getenv("BEFORE_BUILD") or ""
cmd_after_build = os.getenv("AFTER_BUILD") or ""
docker_image_env = os.getenv("DOCKER_IMAGE")
package_site = os.getenv("PACKAGE_SITE") or _DEFAULT_PACKAGE_SITE
minikube_user_root = os.getenv("MINIKUBE_USER_ROOT") or _DEFAULT_MINIKUBE_USER_ROOT
pack_in_cluster = os.getenv("PACK_IN_CLUSTER")

docker_path = os.getenv("DOCKER_PATH")
if docker_path and os.path.isfile(docker_path):
    docker_path = os.path.dirname(docker_path)

_is_linux = sys.platform.lower().startswith("linux")
_is_windows = sys.platform.lower().startswith("win")
_vcs_prefixes = [prefix + "+" for prefix in "git hg svn bzr".split()]

logger = logging.getLogger(__name__)

if sys.version_info[0] == 3:
    unicode = str

dynlibs_pyproject_toml = """
[project]
name = "dynlibs"
version = "0.0.1"

[tool.setuptools.package-data]
dynlibs = ["*.so"]
""".strip()


script_template = r"""
#!/bin/bash
if [[ -n "$NON_DOCKER_MODE" ]]; then
  PYBIN="$(dirname "$PYEXECUTABLE")"
  PIP_PLATFORM_ARGS="--platform=$PYPLATFORM --abi=$PYABI --python-version=$PYVERSION --only-binary=:all:"
  MACHINE_TAG=$("$PYBIN/python" -c "import platform; print(platform.machine())")
  if [[ $(uname) != "Linux" || "$MACHINE_TAG" != "$TARGET_ARCH" ]]; then
    # does not allow compiling under non-linux or different arch
    echo "WARNING: target ($TARGET_ARCH-Linux) not matching host ($MACHINE_TAG-$(uname)), may encounter errors when compiling binary packages."
    PYPI_PLATFORM_INSTALL_ARG="--ignore-requires-python"
    export CC=/dev/null
  fi
else
  PYBIN="/opt/python/{python_abi_version}/bin"
  MACHINE_TAG=$("$PYBIN/python" -c "import platform; print(platform.machine())")
fi
export PATH="$PYBIN:$PATH"

SCRIPT_PATH="$PACK_ROOT/scripts"
BUILD_PKG_PATH="$PACK_ROOT/build"
WHEELS_PATH="$PACK_ROOT/wheels"
INSTALL_PATH="$PACK_ROOT/install"
TEMP_SCRIPT_PATH="$PACK_ROOT/tmp/scripts"
INSTALL_REQUIRE_PATH="$PACK_ROOT/install-req"
WHEELHOUSE_PATH="$PACK_ROOT/wheelhouse"

function handle_sigint() {{
  touch "$SCRIPT_PATH/.cancelled"
  exit 1
}}
trap handle_sigint SIGINT

if [[ "{debug}" == "true" ]]; then
  echo "Files under $SCRIPT_PATH:"
  ls "$SCRIPT_PATH"
  set -e -x
else
  set -e
fi
export PIP_ROOT_USER_ACTION=ignore
export PIP_DISABLE_PIP_VERSION_CHECK=1

if [[ -z "{pypi_index}" ]]; then
  PYPI_EXTRA_ARG=""
else
  PYPI_EXTRA_ARG="-i {pypi_index}"
fi
if [[ "{prefer_binary}" == "true" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --prefer-binary"
fi
if [[ "{use_pep517}" == "true" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --use-pep517"
elif [[ "{use_pep517}" == "false" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --no-use-pep517"
fi
if [[ "{check_build_dependencies}" == "true" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --check-build-dependencies"
fi
if [[ "{no_deps}" == "true" ]]; then
  PYPI_NO_DEPS_ARG="--no-deps"
fi
if [[ "{without_merge}" == "true" ]]; then
  WITHOUT_MERGE="true"
fi
if [[ "{pypi_pre}" == "true" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --pre"
fi
if [[ -n "{pypi_proxy}" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --proxy {pypi_proxy}"
fi
if [[ -n "{pypi_retries}" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --retries {pypi_retries}"
fi
if [[ -n "{pypi_timeout}" ]]; then
  PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --timeout {pypi_timeout}"
fi
if [[ -n "{pypi_trusted_hosts}" ]]; then
  for trusted_host in `echo "{pypi_trusted_hosts}"`; do
    PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --trusted-host $trusted_host"
  done
fi
if [[ -n "{pypi_extra_index_urls}" ]]; then
  for extra_index_url in `echo "{pypi_extra_index_urls}"`; do
    PYPI_EXTRA_ARG="$PYPI_EXTRA_ARG --extra-index-url $extra_index_url"
  done
fi

if [[ -n "$NON_DOCKER_MODE" && "$PYEXECUTABLE" == *"venv"* ]]; then
  VENV_REQS="pip wheel setuptools"
  "$PYBIN/python" -m pip install -U --quiet $PYPI_EXTRA_ARG $VENV_REQS
fi

mkdir -p "$BUILD_PKG_PATH" "$INSTALL_PATH" "$WHEELS_PATH" "$INSTALL_REQUIRE_PATH" "$TEMP_SCRIPT_PATH"

if [[ -f "$SCRIPT_PATH/{_BEFORE_SCRIPT_FILE_NAME}" ]]; then
  echo "Running before build command..."
  source "$SCRIPT_PATH/{_BEFORE_SCRIPT_FILE_NAME}"
  echo ""
fi

if [[ -f "$SCRIPT_PATH/{_INSTALL_REQ_FILE_NAME}" ]]; then
  echo "Installing build prerequisites..."
  "$PYBIN/python" -m pip install --target "$INSTALL_REQUIRE_PATH" \
    -r "$SCRIPT_PATH/{_INSTALL_REQ_FILE_NAME}" $PYPI_EXTRA_ARG
  export OLDPYTHONPATH="$PYTHONPATH"
  export PYTHONPATH="$INSTALL_REQUIRE_PATH:$PYTHONPATH"
  echo ""
fi

# build user-defined package
cd "$BUILD_PKG_PATH"
if [[ -f "$SCRIPT_PATH/requirements-user.txt" ]]; then
  cp "$SCRIPT_PATH/requirements-user.txt" "$TEMP_SCRIPT_PATH/requirements-user.txt"
fi
if [[ -f "$SCRIPT_PATH/{_REQUIREMENT_FILE_NAME}" ]]; then
  cp "$SCRIPT_PATH/{_REQUIREMENT_FILE_NAME}" "$TEMP_SCRIPT_PATH/requirements-extra.txt"
fi

function build_package_at_staging () {{
  mkdir -p "$WHEELS_PATH/staging"
  rm -rf "$WHEELS_PATH/staging/*"

  if [[ -n "$2" ]]; then
    "$PYBIN/python" -m pip wheel --no-deps --wheel-dir "$WHEELS_PATH/staging" $PYPI_EXTRA_ARG "$1" "$2"
  else
    "$PYBIN/python" -m pip wheel --no-deps --wheel-dir "$WHEELS_PATH/staging" $PYPI_EXTRA_ARG "$1"
  fi

  pushd "$WHEELS_PATH/staging" > /dev/null
  for dep_wheel in $(ls *.whl); do
    WHEEL_NAMES="$WHEEL_NAMES $dep_wheel"
    dep_name="$(echo $dep_wheel | sed -E 's/-/ /g' | awk '{{ print $1"=="$2 }}')"
    USER_PACK_NAMES="$USER_PACK_NAMES $dep_name "
    echo "$dep_name" >> "$TEMP_SCRIPT_PATH/requirements-user.txt"
    if [[ -z "$NON_DOCKER_MODE" || "$dep_wheel" == *"-none-"* ]]; then
      mv "$dep_wheel" ../
    fi
  done

  if [[ -z "$PYPI_NO_DEPS_ARG" ]]; then
    cd "$WHEELS_PATH"
    if [[ -z "$NON_DOCKER_MODE" ]]; then
      "$PYBIN/python" -m pip wheel --wheel-dir "$WHEELS_PATH" $PYPI_EXTRA_ARG $WHEEL_NAMES
    else
      "$PYBIN/python" -m pip wheel --wheel-dir "$WHEELS_PATH/staging" $PYPI_EXTRA_ARG --find-links "file://$WHEELS_PATH" $WHEEL_NAMES
      cd "$WHEELS_PATH/staging"
      for dep_wheel in $(ls *.whl); do
        dep_name="$(echo $dep_wheel | sed -E 's/-/ /g' | awk '{{ print $1"=="$2 }}')"
        if [[ "$USER_PACK_NAMES" != *"$dep_name"* ]]; then
          echo "$dep_name" >> "$TEMP_SCRIPT_PATH/requirements-dep-wheels.txt"
          if [[ "$dep_wheel" == *"-none-"* ]]; then
            mv "$dep_wheel" ../
          fi
        fi
      done
    fi
  fi
  popd > /dev/null
}}

echo "Building user-defined packages..."
for path in `ls 2>/dev/null`; do
  if [[ -d "$path" ]]; then
    path="$path/"
  fi
  build_package_at_staging "$path"
done

if [[ -f "$SCRIPT_PATH/{_VCS_FILE_NAME}" ]]; then
  echo ""
  echo "Building VCS packages..."

  if [[ -z "$NON_DOCKER_MODE" ]]; then
    # enable saving password when cloning with git to avoid repeat typing
    git config --global credential.helper store
  fi

  cat "$SCRIPT_PATH/{_VCS_FILE_NAME}" | while read vcs_url ; do
    build_package_at_staging "$vcs_url"
  done
fi

echo ""
echo "Building and installing requirements..."
cd "$WHEELS_PATH"

# download and build all requirements as wheels
if [[ -f "$TEMP_SCRIPT_PATH/requirements-extra.txt" ]]; then
  if [[ -n "$NON_DOCKER_MODE" ]]; then
    last_err_packs=""
    while true; do
      unset BINARY_FAILED
      "$PYBIN/python" -m pip download -r "$TEMP_SCRIPT_PATH/requirements-extra.txt" \
        $PIP_PLATFORM_ARGS $PYPI_EXTRA_ARG $PYPI_NO_DEPS_ARG --find-links "file://$WHEELS_PATH" \
        2> >(tee "$TEMP_SCRIPT_PATH/pip_download_err.log") || export BINARY_FAILED=1
      if [[ -n "$BINARY_FAILED" ]]; then
        # grab malfunctioning dependencies from error message
        err_req_file="$TEMP_SCRIPT_PATH/requirements-downerr.txt"
        cat "$TEMP_SCRIPT_PATH/pip_download_err.log" | grep "matching distribution" \
          | awk '{{print $NF}}' > "$err_req_file"
        err_packs="$(tr '\n' ' ' < "$err_req_file" )"

        # seems we are in an infinite loop
        if [ ! -s "$err_req_file" ] || [[ "$last_err_packs" == "$err_packs" ]]; then
          exit 1
        fi
        last_err_packs="$err_packs"

        echo "Try building pure python wheels $err_packs ..."
        build_package_at_staging -r "$err_req_file"
      else
        break
      fi
    done
  else
    "$PYBIN/python" -m pip wheel -r "$TEMP_SCRIPT_PATH/requirements-extra.txt" $PYPI_EXTRA_ARG $PYPI_NO_DEPS_ARG
  fi
fi

if [[ -f "$TEMP_SCRIPT_PATH/requirements-dep-wheels.txt" ]]; then
  cd "$WHEELS_PATH"
  "$PYBIN/python" -m pip download -r "$TEMP_SCRIPT_PATH/requirements-dep-wheels.txt" \
    $PIP_PLATFORM_ARGS $PYPI_EXTRA_ARG $PYPI_NO_DEPS_ARG
fi

if [[ -n "{dynlibs}" ]] && [[ -z "$PYPI_NO_DEPS_ARG" ]]; then
  echo "Packing configured dynamic binary libraries..."

  mkdir -p "$TEMP_SCRIPT_PATH/dynlibs/dynlibs"
  pushd "$TEMP_SCRIPT_PATH/dynlibs" > /dev/null

  cp "$SCRIPT_PATH/{_DYNLIB_PYPROJECT_TOML_FILE_NAME}" pyproject.toml
  for dynlib in `echo "{dynlibs}"`; do
    if [[ -f "$dynlib" ]]; then
      cp "$dynlib" ./dynlibs 2> /dev/null || true
    else
      for prefix in `echo "/lib /lib64 /usr/lib /usr/lib64 /usr/local/lib /usr/local/lib64 /usr/share/lib /usr/share/lib64"`; do
        cp "$prefix/lib$dynlib"*".so" ./dynlibs 2> /dev/null || true
        cp "$prefix/$dynlib"*".so" ./dynlibs 2> /dev/null || true
      done
    fi
  done

  "$PYBIN/python" -m pip wheel $PYPI_EXTRA_ARG .
  if ls *.whl 1> /dev/null 2>&1; then
    mv *.whl "$WHEELS_PATH/"
    echo "dynlibs" >> "$TEMP_SCRIPT_PATH/requirements-extra.txt"
  fi
  popd > /dev/null
fi

if [[ -n "$(which auditwheel)" ]]; then
  # make sure newly-built binary wheels fixed by auditwheel utility
  for fn in `ls *-linux_$MACHINE_TAG.whl 2>/dev/null`; do
    auditwheel repair "$fn" && rm -f "$fn"
  done
  for fn in `ls dynlibs-*.whl 2>/dev/null`; do
    auditwheel repair "$fn" && rm -f "$fn"
  done
  if [[ -d wheelhouse ]]; then
    mv wheelhouse/*.whl ./
  fi
fi

if [[ -f "$TEMP_SCRIPT_PATH/requirements-user.txt" ]]; then
  cat "$TEMP_SCRIPT_PATH/requirements-user.txt" >> "$TEMP_SCRIPT_PATH/requirements-extra.txt"
fi

export PYTHONPATH="$OLDPYTHONPATH"

if ls "$WHEELS_PATH"/protobuf*.whl 1> /dev/null 2>&1; then
  HAS_PROTOBUF="1"
fi

if [[ -n "$WITHOUT_MERGE" ]]; then
  # move all wheels into wheelhouse
  if [[ -n "$NON_DOCKER_MODE" ]]; then
    mv "$WHEELS_PATH"/*.whl "$WHEELHOUSE_PATH"
  else
    cp --no-preserve=mode,ownership "$WHEELS_PATH"/*.whl "$WHEELHOUSE_PATH"
  fi
  if [[ -f "$SCRIPT_PATH/{_EXCLUDE_FILE_NAME}" ]]; then
    echo ""
    echo "Removing exclusions..."
    for dep in `cat "$SCRIPT_PATH/{_EXCLUDE_FILE_NAME}"`; do
      rm -f "$WHEELHOUSE_PATH/$dep-"*.whl
    done
  fi
else
  # install with recently-built wheels
  "$PYBIN/python" -m pip install --target "$INSTALL_PATH/{package_site}" -r "$TEMP_SCRIPT_PATH/requirements-extra.txt" \
    $PYPI_NO_DEPS_ARG $PIP_PLATFORM_ARGS $PYPI_PLATFORM_INSTALL_ARG --no-index --find-links "file://$WHEELS_PATH"
  rm -rf "$WHEELS_PATH/*"

  if [[ -f "$SCRIPT_PATH/{_EXCLUDE_FILE_NAME}" ]]; then
    echo ""
    echo "Removing exclusions..."

    cd "$INSTALL_PATH/packages"
    for dep in `cat "$SCRIPT_PATH/{_EXCLUDE_FILE_NAME}"`; do
      dist_dir=`ls -d "$dep-"*".dist-info" || echo "non_exist"`
      if [[ ! -f "$dist_dir/RECORD" ]]; then
        continue
      fi
      cat "$dist_dir/RECORD" | while read rec_line ; do
        fn="$(cut -d ',' -f 1 <<< "$rec_line" )"
        cur_root="$(cut -d '/' -f 1 <<< "$fn" )"
        echo "$cur_root" >> /tmp/.rmv_roots
        if [[ -f "$fn" ]]; then
          rm "$fn"
        fi
      done
    done

    if [[ -f "/tmp/.rmv_roots" ]]; then
      for root_dir in `cat /tmp/.rmv_roots | sort | uniq`; do
        find "$root_dir" -type d -empty -delete
      done
    fi
  fi

  # make sure the package is handled as a binary
  touch "$INSTALL_PATH/{package_site}/.pyodps-force-bin.so"

  if [[ "{skip_scan_pkg_resources}" != "true" ]]; then
    echo ""
    echo "Scanning and installing dependency for pkg_resources if needed..."
    if [[ $(egrep --include=\*.py -Rnw "$INSTALL_PATH/{package_site}" -m 1 -e '^\s*(from|import) +pkg_resources' | grep -n 1) ]]; then
      "$PYBIN/python" -m pip install --target "$INSTALL_PATH/{package_site}" \
          $PYPI_NO_DEPS_ARG $PIP_PLATFORM_ARGS setuptools
    else
      echo "No need to install pkg_resources"
    fi
  fi

  echo ""
  echo "Running after build command..."
  {cmd_after_build}

  echo ""
  echo "Packages will be included in your archive:"
  rm -rf "$INSTALL_PATH/{package_site}/dynlibs-0.0.1"* || true  # remove dynlibs package info
  PACKAGE_LIST_FILE="$INSTALL_PATH/{package_site}/.pyodps-pack-meta"
  "$PYBIN/python" -m pip list --path "$INSTALL_PATH/{package_site}" > "$PACKAGE_LIST_FILE"
  cat "$PACKAGE_LIST_FILE"

  echo ""
  echo "Creating archive..."
  mkdir -p "$WHEELHOUSE_PATH"
  cd "$INSTALL_PATH"
  if [[ -n "$MSYSTEM" ]]; then
    pack_path="$(cygpath -u "$WHEELHOUSE_PATH/{_DEFAULT_OUTPUT_FILE}")"
  else
    pack_path="$WHEELHOUSE_PATH/{_DEFAULT_OUTPUT_FILE}"
  fi
  tar --exclude="*.pyc" --exclude="__pycache__" -czf "$pack_path" "{package_site}"

  if [[ -n "$HAS_PROTOBUF" ]]; then
    echo ""
    echo "NOTE: Protobuf detected. You may add \"sys.setdlopenflags(10)\" before using this package in UDFs."
    echo ""
  fi
fi
"""


class PackException(Exception):
    pass


def _indent(text, prefix, predicate=None):
    """Adds 'prefix' to the beginning of selected lines in 'text'.

    If 'predicate' is provided, 'prefix' will only be added to the lines
    where 'predicate(line)' is True. If 'predicate' is not provided,
    it will default to adding 'prefix' to all non-empty lines that do not
    consist solely of whitespace characters.
    """
    """Copied from textwrap.indent method of Python 3"""
    if predicate is None:
        def predicate(line):
            return line.strip()

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)

    return ''.join(prefixed_lines())


def _makedirs(name, mode=0o777, exist_ok=False):
    """makedirs(name [, mode=0o777][, exist_ok=False])

    Super-mkdir; create a leaf directory and all intermediate ones.  Works like
    mkdir, except that any intermediate path segment (not just the rightmost)
    will be created if it does not exist. If the target directory already
    exists, raise an OSError if exist_ok is False. Otherwise no exception is
    raised.  This is recursive.
    """
    """Copied from os.makedirs method of Python 3"""
    head, tail = os.path.split(name)
    if not tail:
        head, tail = os.path.split(head)
    if head and tail and not os.path.exists(head):
        try:
            _makedirs(head, exist_ok=exist_ok)
        except FileExistsError:
            # Defeats race condition when another thread created the path
            pass
        cdir = os.curdir
        if isinstance(tail, bytes):
            cdir = bytes(os.curdir, 'ASCII')
        if tail == cdir:  # xxx/newdir/. exists if xxx/newdir exists
            return
    try:
        os.mkdir(name, mode)
    except OSError:
        # Cannot rely on checking for EEXIST, since the operating system
        # could give priority to other errors like EACCES or EROFS
        if not exist_ok or not os.path.isdir(name):
            raise


def _to_unix(s):
    if isinstance(s, unicode):
        s = s.encode()
    return s.replace(b"\r\n", b"\n")


def _create_build_script(**kwargs):
    template_params = defaultdict(lambda: "")
    template_params.update({k: v for k, v in globals().items() if v is not None})
    template_params.update(kwargs)
    return script_template.lstrip().format(**template_params)


def _copy_to_workdir(src_path, work_dir):
    _makedirs(os.path.join(work_dir, "build"), exist_ok=True)
    path_base_name = os.path.basename(src_path.rstrip("/").rstrip("\\"))
    dest_dir = os.path.join(work_dir, "build", path_base_name)

    shutil.copytree(src_path, dest_dir)


def _copy_package_paths(package_paths=None, work_dir=None, skip_user_path=True):
    remained = []
    for package_path in (package_paths or ()):
        base_name = os.path.basename(package_path.rstrip("/").rstrip("\\"))
        abs_path = os.path.abspath(package_path)
        if not skip_user_path or not abs_path.startswith(os.path.expanduser("~")):
            # not on user path, copy it into build path
            _copy_to_workdir(base_name, work_dir)
        else:
            remained.append(abs_path)
    return remained


def _build_docker_run_command(container_name, docker_image, work_dir, package_paths, docker_args):
    docker_executable = "docker" if not docker_path else os.path.join(docker_path, "docker")
    script_path_mapping = work_dir + "/scripts:/scripts"
    wheelhouse_path_mapping = work_dir + "/wheelhouse:/wheelhouse"
    build_path_mapping = work_dir + "/build:/build"

    cmdline = [docker_executable, "run"]
    if sys.stdin.isatty():
        cmdline.append("-it")

    if docker_args:
        cmdline.extend(shlex.split(docker_args))

    cmdline.extend(["--rm", "--name", container_name])
    cmdline.extend(["-v", script_path_mapping, "-v", wheelhouse_path_mapping])

    if package_paths:
        # need to create build path first for mount
        _makedirs(os.path.join(work_dir, "build"), exist_ok=True)
        cmdline.extend(["-v", build_path_mapping])

    remained = _copy_package_paths(package_paths, work_dir)
    for abs_path in remained:
        base_name = os.path.basename(abs_path.rstrip("/").rstrip("\\"))
        cmdline.extend(["-v", "%s:/build/%s" % (abs_path, base_name)])
    cmdline.extend(
        [docker_image, "/bin/bash", "/scripts/%s" % _PACK_SCRIPT_FILE_NAME]
    )
    return cmdline


def _build_docker_rm_command(container_name):
    docker_executable = "docker" if not docker_path else os.path.join(docker_path, "docker")
    return [docker_executable, "rm", "-f", container_name]


def _log_indent(title, text, indent=2):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(title + "\n%s", _indent(text, " " * indent))


@contextlib.contextmanager
def _create_temp_work_dir(
    requirement_list,
    vcs_list,
    install_requires,
    exclude_list,
    before_script,
    **script_kwargs
):
    try:
        try:
            from pip._vendor.platformdirs import user_cache_dir
        except ImportError:
            from pip._vendor.appdirs import user_cache_dir
        cache_root = user_cache_dir("pyodps-pack")
    except ImportError:
        cache_root = os.path.expanduser("~/.cache/pyodps-pack")

    tmp_path = "%s/pack-root-%d" % (cache_root, int(time.time()))
    try:
        _makedirs(tmp_path, exist_ok=True)
        script_path = os.path.join(tmp_path, "scripts")
        _makedirs(script_path, exist_ok=True)
        _makedirs(os.path.join(tmp_path, "wheelhouse"), exist_ok=True)

        if requirement_list:
            req_text = "\n".join(requirement_list) + "\n"
            _log_indent("Content of requirements.txt:", req_text)
            with open(os.path.join(script_path, _REQUIREMENT_FILE_NAME), "wb") as res_file:
                res_file.write(_to_unix(req_text))

        if vcs_list:
            vcs_text = "\n".join(vcs_list) + "\n"
            _log_indent("Content of requirements-vcs.txt:", vcs_text)
            with open(os.path.join(script_path, _VCS_FILE_NAME), "wb") as res_file:
                res_file.write(_to_unix(vcs_text))

        if install_requires:
            install_req_text = "\n".join(install_requires) + "\n"
            _log_indent("Content of install-requires.txt:", install_req_text)
            with open(os.path.join(script_path, _INSTALL_REQ_FILE_NAME), "wb") as install_req_file:
                install_req_file.write(_to_unix(install_req_text))

        if exclude_list:
            exclude_text = "\n".join(exclude_list) + "\n"
            _log_indent("Content of excludes.txt:", exclude_text)
            with open(os.path.join(script_path, _EXCLUDE_FILE_NAME), "wb") as exclude_file:
                exclude_file.write(_to_unix(exclude_text))

        if before_script or cmd_before_build:
            with open(os.path.join(script_path, _BEFORE_SCRIPT_FILE_NAME), "wb") as before_script_file:
                if before_script:
                    with open(before_script, "rb") as src_before_file:
                        before_script_file.write(_to_unix(src_before_file.read()))
                        before_script_file.write(b"\n\n")
                if cmd_before_build:
                    before_script_file.write(_to_unix(cmd_before_build.encode()))
            if logger.getEffectiveLevel() <= logging.DEBUG:
                with open(os.path.join(script_path, _BEFORE_SCRIPT_FILE_NAME), "r") as before_script_file:
                    _log_indent("Content of before-script.sh:", before_script_file.read())

        with open(os.path.join(script_path, _DYNLIB_PYPROJECT_TOML_FILE_NAME), "wb") as toml_file:
            toml_file.write(_to_unix(dynlibs_pyproject_toml.encode()))

        with open(os.path.join(script_path, _PACK_SCRIPT_FILE_NAME), "wb") as pack_file:
            build_script = _create_build_script(**script_kwargs)
            _log_indent("Pack script:", build_script)
            # make sure script work well under windows
            pack_file.write(_to_unix(build_script.encode()))
        yield tmp_path
    finally:
        if tmp_path and os.path.exists(tmp_path):
            shutil.rmtree(tmp_path)
        try:
            os.rmdir(cache_root)
        except OSError:
            pass


def _get_default_pypi_config():
    def split_config(config_str):
        return [x for x in config_str.strip("'").split("\\n") if x]

    proc = subprocess.Popen(
        [sys.executable, "-m", "pip", "config", "list"], stdout=subprocess.PIPE
    )
    proc.wait()
    if proc.returncode != 0:
        warnings.warn(
            'Failed to call `pip config list`, return code is %s. '
            'Will use default index instead. Specify "-i <index-url>" '
            'if you want to use another package index.' % proc.returncode
        )
        return {}

    sections = defaultdict(dict)
    for line in proc.stdout.read().decode().splitlines():
        var, value = line.split("=", 1)
        if "." not in var:
            continue
        section, var_name = var.split(".", 1)
        sections[section][var_name] = split_config(value)

    items = defaultdict(list)
    for section in ("global", "download", "install"):
        section_values = sections[section]
        for key, val in section_values.items():
            items[key].extend(val)
    return items


def _filter_local_package_paths(parsed_args):
    filtered_req = []
    package_path = []
    vcs_urls = []

    parsed_args.specifiers = list(parsed_args.specifiers)
    for req_file in parsed_args.requirement:
        with open(req_file, "r") as input_req_file:
            for req in input_req_file:
                parsed_args.specifiers.append(req)

    for req in parsed_args.specifiers:
        if re.findall(r"[^=\>\<]=[^=\>\<]", req):
            req = req.replace("=", "==")
        if os.path.exists(req):
            package_path.append(req)
        elif any(req.startswith(prefix) for prefix in _vcs_prefixes):
            vcs_urls.append(req)
        else:
            filtered_req.append(req)
    parsed_args.specifiers = filtered_req
    parsed_args.package_path = package_path
    parsed_args.vcs_urls = vcs_urls


def _collect_install_requires(parsed_args):
    install_requires = parsed_args.install_requires or []
    for req_file_name in parsed_args.install_requires_file:
        with open(req_file_name, "r") as req_file:
            install_requires.extend(req_file.read().splitlines())
    parsed_args.install_requires = install_requires


def _collect_env_packages(exclude_editable=False, exclude=None, index_url=None):
    print("Extracting packages from local environment...")
    exclude = set(exclude or [])
    pip_cmd = [sys.executable, "-m", "pip", "list", "--format", "json"]
    if exclude_editable:
        pip_cmd += ["--exclude-editable"]
    if index_url:
        pip_cmd += ["--index-url", index_url]

    pack_descriptions = []
    proc = subprocess.Popen(pip_cmd, stdout=subprocess.PIPE)
    proc.wait()
    if proc.returncode != 0:
        raise PackException(
            "ERROR: Failed to call `pip list`. Return code is %s" % proc.returncode
        )
    pack_descriptions.extend(json.loads(proc.stdout.read()))

    specifiers = []
    missing_packs = []
    for desc in pack_descriptions:
        pack_name = desc["name"]
        if pack_name in exclude:
            continue
        if "editable_project_location" in desc:
            specifiers.append(desc["editable_project_location"])
        else:
            specifiers.append("%s==%s" % (desc["name"], desc["version"]))
    if missing_packs:
        warnings.warn(
            "Cannot find packages %s in package index. These packages cannot be included."
            % ",".join(missing_packs)
        )
    return specifiers


def _get_arch(arch=None):
    arch = (arch or "x86_64").lower()
    if arch in ("arm64", "aarch64"):
        return "aarch64"
    elif arch == "x86_64":
        return arch
    raise PackException("Arch %s not supported" % arch)


def _get_default_image(use_legacy_image=False, arch=None):
    arch = _get_arch(arch)
    if arch != "x86_64" and use_legacy_image:
        raise PackException("Cannot use legacy image when building on other arches")
    if use_legacy_image:
        return _LEGACY_DOCKER_IMAGE_X64
    elif arch == "x86_64":
        return _DEFAULT_DOCKER_IMAGE_X64
    elif arch == "aarch64":
        return _DEFAULT_DOCKER_IMAGE_ARM64
    else:
        raise PackException("Arch %s not supported" % arch)


def _get_python_abi_version(python_version=None, mcpy27=None, dwpy27=None):
    if python_abi_env and python_version is not None:
        raise PackException(
            "You should not specify environment variable 'PYABI' and '--python-version' at the same time."
        )
    if python_version is None:
        python_abi_version = python_abi_env or _DEFAULT_PYABI
    else:
        if "." not in python_version:
            version_parts = (int(python_version[0]), int(python_version[1:]))
        else:
            version_parts = tuple(int(pt) for pt in python_version.split("."))[:2]
        cp_tag = "cp%d%d" % version_parts
        python_abi_version = cp_tag + "-" + cp_tag
        if version_parts < (3, 8):
            python_abi_version += "m"
    if dwpy27:
        if mcpy27:
            raise PackException("You should not specify '--dwpy27' and '--mcpy27' at the same time.")
        python_abi_version = _DWPY27_PYABI
    elif mcpy27:
        python_abi_version = _MCPY27_PYABI
    return python_abi_version


def _get_bash_path():
    """Get bash executable. When under Windows, retrieves path of Git bash. Otherwise returns /bin/bash."""
    if not _is_windows:
        return "/bin/bash"

    import winreg

    key = None
    try:
        key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\GitForWindows")
        git_path = winreg.QueryValueEx(key, "InstallPath")[0]
        bash_path = os.path.join(git_path, "bin", "bash.exe")
        if not os.path.exists(bash_path):
            raise OSError("bash.exe not found")
        return bash_path
    except OSError:
        err_msg = (
            "Failed to locate Git Bash. Please check your installation of Git for Windows "
            "which can be obtained at https://gitforwindows.org/."
        )
        if int(platform.win32_ver()[1].rsplit(".")[-1]) > 19041:
            err_msg += " You may also try packing under WSL or with Docker."
        else:
            err_msg += " You may also try packing with Docker."
        raise PackException(err_msg)
    finally:
        if key:
            key.Close()


def _get_local_pack_executable(work_dir):
    """Create a virtualenv for local packing if possible"""
    try:
        import venv
    except ImportError:
        return sys.executable

    env_dir = os.path.join(work_dir, "venv")
    print("Creating virtual environment for local build...")
    venv.create(env_dir, symlinks=not _is_windows, with_pip=True)
    if _is_windows:
        return os.path.join(env_dir, "Scripts", "python.exe")
    else:
        return os.path.join(env_dir, "bin/python")


def _rewrite_minikube_command(docker_cmd, done_timeout=10):
    if "MINIKUBE_ACTIVE_DOCKERD" not in os.environ:
        return docker_cmd, None

    print("Mounting home directory to minikube...")
    user_home = os.path.expanduser("~")
    mount_cmd = ["minikube", "mount", user_home + ":" + minikube_user_root]
    logger.debug("Minikube mount command: %r", mount_cmd)
    proc = subprocess.Popen(mount_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    mount_event = threading.Event()

    def wait_start_thread_func():
        while True:
            line = proc.stdout.readline()
            if not line and proc.poll() is not None:
                break
            try:
                line_str = line.decode("utf-8").rstrip()
                logger.debug(u"Output of minikube mount: %s", line_str)
            except:
                logger.debug("Output of minikube mount: %r", line)
            if b"mounted" in line:
                mount_event.set()

    def rewrite_docker_cmd_part(part):
        if user_home in part:
            part = part.replace(user_home, minikube_user_root)
            return part.replace(os.path.sep, "/")
        return part

    wait_thread = threading.Thread(target=wait_start_thread_func)
    wait_thread.start()
    if not mount_event.wait(done_timeout):
        raise PackException("Mount minikube directory timed out.")

    new_docker_cmd = [rewrite_docker_cmd_part(part) for part in docker_cmd]
    return new_docker_cmd, proc


def _main(parsed_args):
    if parsed_args.debug:
        logging.basicConfig(level=logging.DEBUG)

    if parsed_args.pack_env:
        if parsed_args.specifiers:
            raise PackException(
                "ERROR: Cannot supply --pack-env with other package specifiers."
            )
        parsed_args.specifiers = _collect_env_packages(
            parsed_args.exclude_editable, parsed_args.exclude, parsed_args.index_url
        )

    _filter_local_package_paths(parsed_args)
    _collect_install_requires(parsed_args)

    if not parsed_args.specifiers and not parsed_args.package_path and not parsed_args.vcs_urls:
        raise PackException("ERROR: You must give at least one requirement to install.")

    file_cfg = _get_default_pypi_config()

    def _first_or_none(list_val):
        return list_val[0] if list_val else None

    index_url = parsed_args.index_url or _first_or_none(file_cfg.get("index-url")) or ""

    if index_url:
        logger.debug("Using PyPI index %s", index_url)
    else:
        logger.debug("Using default PyPI index")

    prefer_binary_str = "true" if parsed_args.prefer_binary else ""
    no_deps_str = "true" if parsed_args.no_deps else ""
    debug_str = "true" if parsed_args.debug else ""
    without_merge_str = "true" if parsed_args.without_merge else ""
    use_pep517_str = str(parsed_args.use_pep517).lower()
    check_build_dependencies_str = "true" if parsed_args.check_build_dependencies else ""
    skip_scan_pkg_resources_str = "true" if parsed_args.skip_scan_pkg_resources else ""
    pre_str = "true" if parsed_args.pre else ""
    timeout_str = parsed_args.timeout or _first_or_none(file_cfg.get("timeout")) or ""
    proxy_str = parsed_args.proxy or _first_or_none(file_cfg.get("proxy")) or ""
    retries_str = parsed_args.retries or _first_or_none(file_cfg.get("retries")) or ""
    dynlibs_str = " ".join(parsed_args.dynlib)

    python_abi_version = _get_python_abi_version(
        parsed_args.python_version, parsed_args.mcpy27, parsed_args.dwpy27
    )

    extra_index_urls = (parsed_args.extra_index_url or []) + (
        file_cfg.get("extra-index-url") or []
    )
    extra_index_urls_str = " ".join(extra_index_urls)

    trusted_hosts = (parsed_args.trusted_host or []) + (
        file_cfg.get("trusted-host") or []
    )
    trusted_hosts_str = " ".join(trusted_hosts)

    with _create_temp_work_dir(
        parsed_args.specifiers,
        parsed_args.vcs_urls,
        parsed_args.install_requires,
        parsed_args.exclude,
        parsed_args.run_before,
        pypi_pre=pre_str,
        pypi_index=index_url,
        pypi_extra_index_urls=extra_index_urls_str,
        pypi_proxy=proxy_str,
        pypi_retries=retries_str,
        pypi_timeout=timeout_str,
        prefer_binary=prefer_binary_str,
        use_pep517=use_pep517_str,
        check_build_dependencies=check_build_dependencies_str,
        skip_scan_pkg_resources=skip_scan_pkg_resources_str,
        no_deps=no_deps_str,
        without_merge=without_merge_str,
        python_abi_version=python_abi_version,
        pypi_trusted_hosts=trusted_hosts_str,
        dynlibs=dynlibs_str,
        debug=debug_str,
    ) as work_dir:
        container_name = "pack-cnt-%d" % int(time.time())

        use_legacy_image = parsed_args.legacy_image or parsed_args.mcpy27 or parsed_args.dwpy27
        default_image = _get_default_image(use_legacy_image, parsed_args.arch)
        docker_image = docker_image_env or default_image

        minikube_mount_proc = None
        if pack_in_cluster or parsed_args.without_docker:
            _copy_package_paths(parsed_args.package_path, work_dir, skip_user_path=False)

            pyversion, pyabi = python_abi_version.split("-", 1)
            pyversion = pyversion[2:]
            build_cmd = [_get_bash_path(), os.path.join(work_dir, "scripts", _PACK_SCRIPT_FILE_NAME)]
            build_env = {
                "PACK_ROOT": work_dir,
                "PYPLATFORM": default_image.replace("quay.io/pypa/", ""),
                "PYVERSION": pyversion,
                "PYABI": pyabi,
                "TARGET_ARCH": _get_arch(parsed_args.arch),
            }
            if parsed_args.without_docker:
                build_env["NON_DOCKER_MODE"] = "true"
                build_env["PYEXECUTABLE"] = _get_local_pack_executable(work_dir)
            else:
                temp_env = build_env
                build_env = os.environ.copy()
                build_env.update(temp_env)
                build_env["PACK_IN_CLUSTER"] = "true"
            logger.debug("Command: %r", build_cmd)
            logger.debug("Environment variables: %r", build_cmd)
        else:
            build_cmd = _build_docker_run_command(
                container_name,
                docker_image,
                work_dir,
                parsed_args.package_path,
                parsed_args.docker_args,
            )
            build_cmd, minikube_mount_proc = _rewrite_minikube_command(build_cmd)
            build_env = None
            logger.debug("Docker command: %r", build_cmd)

        try:
            proc = subprocess.Popen(build_cmd, env=build_env)
        except OSError as ex:
            if ex.errno != errno.ENOENT:
                raise

            logger.error("Failed to execute command %r, the error message is %s.", build_cmd, ex)
            if pack_in_cluster or parsed_args.without_docker:
                if _is_windows:
                    raise PackException(
                        "Cannot locate git bash. Please install Git for Windows or "
                        "try WSL instead."
                    )
                else:
                    # in MacOS or Linux, this error is not a FAQ, thus just raise it
                    raise
            else:
                raise PackException(
                    "Cannot locate docker. Please install it, reopen your terminal and "
                    "retry. Or you may try `--without-docker` instead. If you've already "
                    "installed Docker, you may specify the path of its executable via "
                    "DOCKER_PATH environment."
                )
        cancelled = False
        try:
            proc.wait()
        except KeyboardInterrupt:
            cancelled = True
            if not parsed_args.without_docker and not pack_in_cluster:
                docker_rm_cmd = _build_docker_rm_command(container_name)
                logger.debug("Docker rm command: %r", docker_rm_cmd)
                subprocess.Popen(docker_rm_cmd, stdout=subprocess.PIPE)
                proc.wait()
        finally:
            if minikube_mount_proc is not None:
                minikube_mount_proc.terminate()

        if proc.returncode != 0:
            cancelled = cancelled or os.path.exists(os.path.join(work_dir, "scripts", ".cancelled"))
            if cancelled:
                print("Cancelled by user.")
            else:
                if parsed_args.without_docker:
                    print(
                        "Errors occurred when creating your package. This is often caused "
                        "by mismatching Python version, platform or architecture when "
                        "encountering binary packages. Please check outputs for details. "
                        "You may try building your packages inside Docker by removing "
                        "--without-docker option, which often resolves the issue."
                    )
                else:
                    print(
                        "Errors occurred when creating your package. Please check outputs "
                        "for details. You may add a `--debug` option to obtain more "
                        "information."
                    )

                if proc.returncode == _SEGFAULT_ERR_CODE and use_legacy_image:
                    print(
                        "Image manylinux1 might crash silently under some Docker environments. "
                        "You may try under a native Linux environment. Details can be seen at "
                        "https://mail.python.org/pipermail/wheel-builders/2016-December/000239.html."
                    )
                elif _is_linux and "SUDO_USER" not in os.environ:
                    print(
                        "You need to run pyodps-pack with sudo to make sure docker is "
                        "executed properly."
                    )
        else:
            if parsed_args.without_merge:
                src_path = os.path.join(work_dir, "wheelhouse", "*.whl")
                for wheel_name in glob.glob(src_path):
                    shutil.move(wheel_name, os.path.basename(wheel_name))
            else:
                src_path = os.path.join(work_dir, "wheelhouse", _DEFAULT_OUTPUT_FILE)
                shutil.move(src_path, parsed_args.output)

            if _is_linux and "SUDO_UID" in os.environ and "SUDO_GID" in os.environ:
                own_desc = "%s:%s" % (os.environ["SUDO_UID"], os.environ["SUDO_GID"])
                target_path = "*.whl" if parsed_args.without_merge else parsed_args.output
                chown_proc = subprocess.Popen(["chown", own_desc, target_path])
                chown_proc.wait()

            if parsed_args.without_merge:
                print("Result wheels stored at current dir")
            else:
                print("Result package stored as %s" % parsed_args.output)
        return proc.returncode


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "specifiers", metavar="REQ", nargs="*",
        help="a requirement item compatible with pip command",
    )
    parser.add_argument(
        "--requirement", "-r", action="append", default=[],
        metavar="PATH", help="Path of requirements.txt including file name",
    )
    parser.add_argument(
        "--install-requires", action="append", default=[], help="Requirement for install time"
    )
    parser.add_argument(
        "--install-requires-file", action="append", default=[],
        help="Requirement file for install time",
    )
    parser.add_argument(
        "--run-before", help="Prepare script before package build."
    )
    parser.add_argument(
        "--no-deps", action="store_true", default=False,
        help="Don't put package dependencies into archives",
    )
    parser.add_argument(
        "--pre", action="store_true", default=False,
        help="Include pre-release and development versions. "
             "By default, pyodps-pack only finds stable versions.",
    )
    parser.add_argument(
        "--proxy", metavar="proxy",
        help="Specify a proxy in the form scheme://[user:passwd@]proxy.server:port.",
    )
    parser.add_argument(
        "--retries", metavar="retries",
        help="Maximum number of retries each connection should attempt (default 5 times).",
    )
    parser.add_argument(
        "--timeout", metavar="sec", help="Set the socket timeout (default 15 seconds).",
    )
    parser.add_argument(
        "--exclude", "-X", action="append", default=[],
        metavar="DEPEND", help="Requirements to exclude from the package",
    )
    parser.add_argument(
        "--index-url", "-i", default="",
        help="Base URL of PyPI package. If absent, will use "
             "`global.index-url` in `pip config list` command by default.",
    )
    parser.add_argument(
        "--extra-index-url", metavar="url", action="append", default=[],
        help="Extra URLs of package indexes to use in addition to --index-url. "
             "Should follow the same rules as --index-url.",
    )
    parser.add_argument(
        "--trusted-host", metavar="host", action="append", default=[],
        help="Mark this host or host:port pair as trusted, "
             "even though it does not have valid or any HTTPS.",
    )
    parser.add_argument(
        "--legacy-image", "-l", action="store_true", default=False,
        help="Use legacy image to make packages",
    )
    parser.add_argument(
        "--mcpy27", action="store_true", default=False,
        help="Build package for Python 2.7 on MaxCompute. "
             "If enabled, will assume `legacy-image` to be true.",
    )
    parser.add_argument(
        "--dwpy27", action="store_true", default=False,
        help="Build package for Python 2.7 on DataWorks. "
             "If enabled, will assume `legacy-image` to be true.",
    )
    parser.add_argument(
        "--prefer-binary", action="store_true", default=False,
        help="Prefer older binary packages over newer source packages",
    )
    parser.add_argument(
        "--output", "-o", default="packages.tar.gz", help="Target archive file name to store"
    )
    parser.add_argument(
        "--dynlib", action="append", default=[],
        help="Dynamic library to include. Can be an absolute path to a .so library "
             "or library name with or without 'lib' prefix."
    )
    parser.add_argument(
        "--pack-env", action="store_true", default=False, help="Pack full environment"
    )
    parser.add_argument(
        "--exclude-editable", action="store_true", default=False,
        help="Exclude editable packages when packing",
    )
    parser.add_argument(
        "--use-pep517", action="store_true", default=None,
        help="Use PEP 517 for building source distributions (use --no-use-pep517 to force legacy behaviour).",
    )
    parser.add_argument(
        "--no-use-pep517", action="store_false", dest="use_pep517", default=None, help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--check-build-dependencies", action="store_true", default=None,
        help="Check the build dependencies when PEP517 is used.",
    )
    parser.add_argument(
        "--arch", default="x86_64",
        help="Architecture of target package, x86_64 by default. Currently only x86_64 "
             "and aarch64 supported. Do not use this argument if you are not running "
             "your code in a proprietary cloud."
    )
    parser.add_argument(
        "--python-version",
        help="Version of Python your environment is on, for instance 3.6. "
             "You may also use 36 instead. Do not use this argument if you "
             "are not running your code in a proprietary cloud."
    )
    parser.add_argument(
        "--docker-args", help="Extra arguments for Docker."
    )
    parser.add_argument(
        "--without-docker", action="store_true", default=False,
        help="Create packages without Docker. May cause errors if incompatible "
             "binaries involved.",
    )
    parser.add_argument(
        "--without-merge", action="store_true", default=False,
        help="Create or download wheels without merging them.",
    )
    parser.add_argument(
        "--skip-scan-pkg-resources", action="store_true", default=False,
        help="Skip scanning for usage of pkg-resources package.",
    )
    parser.add_argument(
        "--debug", action="store_true", default=False,
        help="Dump debug messages for diagnose purpose",
    )

    args = parser.parse_args()

    try:
        sys.exit(_main(args) or 0)
    except PackException as ex:
        print(ex.args[0], file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from mars.deploy.kubernetes.worker import K8SWorkerCommandRunner

from .core import CupidCommandRunnerMixin

logger = logging.getLogger(__name__)


class CupidWorkerCommandRunner(CupidCommandRunnerMixin, K8SWorkerCommandRunner):
    def __call__(self, *args, **kwargs):
        try:
            self.fix_protobuf_import()
            self.start_cupid_service()
            super().__call__(*args, **kwargs)
        except:
            import traceback

            traceback.print_exc()
            raise
        finally:
            self.stop_cupid_service()

    def parse_args(self, parser, argv, environ=None):
        self.fix_hyper_address()
        return super().parse_args(parser, argv, environ)

    async def start_services(self):
        await self.write_node_endpoint()
        await super().start_services()


main = CupidWorkerCommandRunner()

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *
from libcpp.vector cimport vector

from ...src.types_c cimport SchemaSnapshot
from ..checksum_c cimport Checksum
from .util_c cimport *
from .block_decoder_c cimport Decoder


ctypedef void (*_NOGIL_READER)(TunnelPandasReader self, ArrayVariantPtrs &record, int row) nogil


cdef class TunnelPandasReader:
    cdef object _schema
    cdef object _columns
    cdef object _stream
    cdef Checksum _crc
    cdef Checksum _crccrc
    cdef int _n_columns
    cdef int _read_limit
    cdef object _reader_schema
    cdef SchemaSnapshot _schema_snapshot
    cdef Decoder _decoder

    cdef int _cur_cursor
    cdef int _mem_cache_size
    cdef int _mem_cache_bound
    cdef int _row_cache_size

    cdef vector[_NOGIL_READER] _nogil_readers
    cdef object _mem_cache
    cdef int _row_mem_ptr
    cdef uint32_t _row_checksum
    cdef int _row_ptr
    cdef int _use_no_gil

    cdef void _read_bool(self, ArrayVariantPtrs &aptr, int row) nogil
    cdef void _read_int64(self, ArrayVariantPtrs &aptr, int ptr) nogil
    cdef void _read_float(self, ArrayVariantPtrs &aptr, int idx) nogil
    cdef void _read_double(self, ArrayVariantPtrs &aptr, int row) nogil
    cdef int _fill_ndarrays_nogil(self, vector[ArrayVariantPtrs] &col_ptrs, int start_row, int limit) nogil

    cdef void _scan_schema(self)
    cpdef refill_cache(self)
    cpdef reset_positions(self, object cache, int cache_size)

    cpdef int readinto(self, object buffers, object columns=*, int limit=*) except? -1
